from libtad.authentication import Authentication
import libtad.constants as Constants
from typing import Dict, List

class BaseService:
    def __init__(self, access_key: str, secret_key: str, service_name: str):
        self._version: int = 3
        self.__language: List[str] = [Constants.DEFAULTLANGUAGE]
        self._service_name: str = service_name
        auth = Authentication(service_name, access_key, secret_key)
        self._authentication_options: Dict[str, object] = auth.get_authentication_args()

    @property
    def language(self):
        return self.__language

    @language.setter
    def language(self, value):
        if type(value) is str:
            self.__language = [value]
        else:
            self.__language = value




import libtad.constants as Constants
from libtad.base_service import BaseService
from libtad.datatypes.business import BusinessDaysOperatorType, BusinessDaysFilterType, BusinessDates
from libtad.datatypes.places import LocationId
from libtad.datatypes.time import TADDateTime
from libtad.common import XmlUtils
import xml.etree.ElementTree as ET
from urllib.parse import urlencode
from urllib.request import urlopen, Request
from ssl import SSLContext
from typing import List, Dict, Union

class BusinessDateService(BaseService):
    """
    The businessdate service can be used to find a business date from a 
    specified number of days. 

    You can choose if you want to calculate the business date by 
    adding (default) or subtracting the given days, and whether or not a 
    specific filter should be applied to the result. By default the result 
    will be filtered on excluding weekends and public holidays, but you 
    can specify a custom filter to modify this.   

    ...

    Attributes
    ----------
    include : bool
        Specify whether the result should be calculated by including instead 
        of excluding the days.
    
    filter : BusinessDaysFilterType or list of BusinessDaysFilterType
        Choose a set of types or days you want to filter on.

    operator : BusinessDaysOperatorType
        Set if the service should do an addition or subtraction of the 
        specified days.

    repeat : int
        Set how many times the calculation should be repeated (only applicable 
        when days parameter has exactly one number).

    language : list of str
        The preferred language(s) for the texts. An error will be raised if the
        language code cannot be recognized. In case the text for a specific event
        cannot be retrieved in any of the requested languages it will be returned
        in English instead.

        It is possible to query all available languages in a single request by
        specifying a primary text language, followed by the special value _all_.
        This value can not be specified as the first language.

        Example:
        ``service.language = ["de"]``

        Example:
        ``service.language.append("de")``

    Methods
    -------
    get_business_date_for_place(place_id, start_date, days)
        Gets the business dates by place id.   
    get_business_date_for_country(country_iso, start_date, days, state_iso=None)
        Gets the business dates by country and an optional state.   
    """

    def __init__(self, access_key: str, secret_key: str):
        """
        Parameters
        ----------
        access_key : str
            Access key.
        secret_key : str
            Secret key.
        """

        super().__init__(access_key, secret_key, "businessdate")
        self.include: bool = False
        self.filter: Union[BusinessDaysFilterType, List[BusinessDaysFilterType]] = BusinessDaysFilterType.Weekendholidays
        self.operator: BusinessDaysOperatorType = BusinessDaysOperatorType.Add
        self.repeat: int = 1

    def get_business_date_for_place(self, place_id: LocationId, start_date: TADDateTime, days: Union[int, List[int]]) -> BusinessDates:
        """
        Gets the business dates by place id.   

        Parameters
        ----------
        place_id : LocationId
            Place identifier.
        start_date : TADDateTime
            The first date to count.
        days : int or list of int
            How many business days to count.

        Returns
        -------
        business_dates : BusinessDates
            Object containing geographical information and a list of the 
            calculated result for the requested period.

        """

        if not isinstance(place_id, LocationId):
            raise ValueError("An argument is invalid")

        ID = str(place_id)
        if not ID:
            raise ValueError("An argument is invalid")

        args: Dict[str, object] = {"placeid": ID}
        return self.__get_business_date(args, start_date, days)

    def get_business_date_for_country(self, country_iso: str, start_date: TADDateTime, days: Union[int, List[int]], state_iso: str = None) -> BusinessDates:
        """
        Gets the business dates by country and an optional state.   

        Parameters
        ----------
        country_iso : LocationId
            The country for which you would like to calculate the business date.
        start_date : TADDateTime
            The first date to count.
        days : int or list of int
            How many business days to count.
        state_iso : str, optional
            The state in the given country you want to calculate the business date.
            Uses no states as default.

        Returns
        -------
        business_dates : BusinessDates
            Object containing geographical information and a list of the 
            calculated result for the requested period.

        """

        if not (isinstance(country_iso, str) and (not state_iso or isinstance(state_iso, str))):
            raise ValueError("An argument is invalid")

        args: Dict[str, object] = {"country": country_iso}
        if (state_iso):
            args["state"] = state_iso

        return self.__get_business_date(args, start_date, days)

    def __get_business_date(self, args: Dict[str, object], start_date: TADDateTime, days: Union[int, List[int]]) -> BusinessDates:
        if not isinstance(start_date, TADDateTime):
            raise ValueError("An argument is invalid")

        if isinstance(days, int):
            days_str = str(days)
        elif isinstance(days, list) and all(isinstance(day, int) for day in days):
            days_str = ",".join(map(str, days))
        else:
            raise ValueError("An argument is invalid")

        if isinstance(days, int) or len(days) == 1:
            args["repeat"] = self.repeat

        args.update(self._authentication_options)
        args["startdt"] = str(start_date)
        args["days"] = days_str

        return self.__retrieve_business_date(args)

    def __retrieve_business_date(self, args: Dict[str, object]) -> BusinessDates:
        arguments: Dict[str, object] = self.__get_optional_arguments(args)
        url: str = Constants.ENTRYPOINT + "/" + self._service_name + "?" + urlencode(arguments)
        req = Request(
                url,
                headers = {"User-Agent": "libtad-py"}
                )

        with urlopen(req, context=SSLContext()) as f:
            result: str = f.read().decode("utf-8")

        return self.__from_xml(result)

    def __get_optional_arguments(self, args: Dict[str, object]) -> Dict[str, object]:
        optional_args: Dict[str, str] = args

        optional_args["include"] = int(self.include)
        optional_args["lang"] = ",".join(self.language)
        optional_args["out"] = Constants.DEFAULTRETURNFORMAT
        optional_args["verbosetime"] = str(Constants.DEFAULTVERBOSETIMEVALUE)
        optional_args["version"] = str(self._version)

        if isinstance(self.filter, BusinessDaysFilterType):
            optional_args["filter"] = str(self.filter)
        elif isinstance(self.filter, list) and all(isinstance(filter, BusinessDaysFilterType) for filter in self.filter):
            optional_args["filter"] = ",".join(map(str, self.filter))

        if isinstance(self.operator, BusinessDaysOperatorType):
            optional_args["op"] = self.operator.name.lower()

        return optional_args

    def __from_xml(self, result: str) -> BusinessDates:
        XmlUtils.check_for_errors(result)
        xml: ET.Element = ET.fromstring(result)
        return BusinessDates(xml.find("geo"), [period for period in xml.findall("period")])




from .coordinates import Coordinates
from .place import Place
from typing import Union

class LocationId:
    """
    LocationId class represents the different kinds of IDs that Time and Date API
    does support. 

    ...

    Attributes
    ----------
    textual_id : str
        Can be country code, country name, city, etc.
    numeric_id : int
        Usually an integer ID that is returned from a previous API call.
    coordinates : Coordinates
        Provide an coordinate object to LocationId.
    """

    def __init__(self, location: Union[str, int, Coordinates, Place]):
        self.__textual_id: str = None
        self.numeric_id: int = None
        self.__coordinates: Coordinates = None
        
        if type(location) is str:
            self.__textual_id = location
        elif type(location) is int:
            self.numeric_id = location
        elif type(location) is Coordinates:
            self.__coordinates = location
        elif type(location) is Place:
            self.numeric_id = location.id
        else:
            raise ValueError("Location argument must be of type str, int, Coordinates or Place");

    @property
    def textual_id(self) -> str:
        return self.__textual_id

    @property
    def coordinates(self) -> Coordinates:
        return self.__coordinates

    def __str__(self):
        if self.numeric_id is not None:
            return str(self.numeric_id)
        elif self.__textual_id is not None:
            return self.__textual_id
        elif self.__coordinates is not None:
            return str(self.__coordinates)
        else:
            return ""




from .tad_date_time import TADDateTime
from .tad_timezone import TADTimezone
import xml.etree.ElementTree as ET

class TADTime:
    '''
    A class used to store information about time and timezone.

    ...

    Attributes
    ----------
    iso : str
        ISO representation of date and time, timezone included if
        different from UTC. If time is not applicable, only the
        date is shown.

        Example: 2011-06-08T09:18:16+02:00

        Example: 2011-06-08T07:18:16 (UTC time)

        Example: 2011-06-08 (only date)

    datetime : TADDateTime
        Date and time representation of the ISO string.

    timezone : TADTimezone
        Timezone information. Element is only present if different
        from UTC and requested by specifying the
        IncludeTimezoneInformation parameter.
    '''

    def __init__(self, node: ET.Element):
        self.iso: str = ""
        self.datetime: TADDateTime = None
        self.timezone: TADTimezone = None
        
        iso = node.get("iso")
        timezone = node.find("timezone")
        datetime = node.find("datetime")

        if iso:
            self.iso = iso

        if timezone is not None:
            self.timezone = TADTimezone(timezone)

        if datetime is not None:
            year_node = datetime.find("year")
            month_node = datetime.find("month")
            day_node = datetime.find("day")
            hour_node = datetime.find("hour")
            minute_node = datetime.find("minute")
            second_node = datetime.find("second")

            year = int(year_node.text) if year_node is not None else 0
            month = int(month_node.text) if month_node is not None else 1
            day = int(day_node.text) if day_node is not None else 1
            hour = int(hour_node.text) if hour_node is not None else 0
            minute = int(minute_node.text) if minute_node is not None else 0
            second = int(second_node.text) if second_node is not None else 0

            self.datetime = TADDateTime(year, month, day, hour, minute, second, 0)
        elif self.iso:
            self.datetime = TADDateTime._parse(self.iso)




from .geo import Geo
import xml.etree.ElementTree as ET

class Place:
    """
    A class used for places

    ...

    Attributes
    ----------
    id : int
        Numerical id of the referenced place.
    urlid : str
        Textual id of the referenced place.
    geography : Geo
        Geographical information about the location.
    """

    def __init__(self, node: ET.Element):
        self.id: int = None
        self.urlid: str = ""
        self.geography: Geo = None

        ID = node.get("id")
        urlid = node.get("urlid")
        geo = node.find("geo")

        if ID:
            self.id = ID

        if urlid:
            self.urlid = urlid

        if geo is not None:
            self.geography = Geo(geo)




from enum import Enum

class BusinessDaysFilterType(Enum):
    """
    An enum class containing business.PeriodTypeday filters.

    ...

    Attributes
    ----------
    Mon : BusinessDaysFilterType
        Monday.
    Tue : BusinessDaysFilterType
        Tuesday.
    Wed : BusinessDaysFilterType
        Wednesday.
    Thu : BusinessDaysFilterType
        Thursday.
    Fri : BusinessDaysFilterType
        Friday.
    Sat : BusinessDaysFilterType
        Saturday.
    Sun : BusinessDaysFilterType
        Sunday.
    All : BusinessDaysFilterType
        All days.
    Weekend : BusinessDaysFilterType
        Weekend.
    Holidays : BusinessDaysFilterType
        Holidays.
    Weekendholidays : BusinessDaysFilterType
        Weekend and holidays.
    Nothing : BusinessDaysFilterType
        No days.
    """

    Mon = 1
    Tue = 2
    Wed = 3
    Thu = 4
    Fri = 5
    Sat = 6
    Sun = 7
    All = 8
    Weekend = 9
    Holidays = 10
    Weekendholidays = 11
    Nothing = 12

    def __str__(self):
        if self == BusinessDaysFilterType.Nothing:
            return "none"
        else:
            return self.name.lower()




from .astronomy_event_type import AstronomyEventType
from libtad.common.exceptions import MalformedXMLException
import xml.etree.ElementTree as ET

class AstronomyEvent:
    """
    Contains sunrise/sunset event of a given day.

    ...

    Attributes
    ----------
    type : AstronomyEventType
        Indicates the type of the event.
    hour : int
        Hour at which the event is happening (local time).
    minute : int
        minute at which the event is happening (local time).
    """

    def __init__(self, node: ET.Element):
        self.type: AstronomyEventType = None
        self.hour: int = None
        self.minute: int = None

        type_ = node.get("type")
        hour = node.get("hour")
        minute = node.get("minute")

        if type_:
            try:
                self.type = AstronomyEventType[type_.capitalize()]
            except KeyError:
                pass

        if not self.type:
            raise MalformedXMLException("The XML Received from Time and Date did not "
                + "include an event type which complies with an AstronomyEventType enum: " + type_, ignore_default=True)

        if hour and hour.isdecimal():
            self.hour = int(hour)

        if minute and minute.isdecimal():
            self.minute = int(minute)




from enum import Flag

class AstronomyEventCode(Flag):
    """
    An enum class containing astronomy events.

    ...

    Attributes
    ----------
    AstronomicalTwilightStarts : AstronomyEventCode
        Astronomical twilight (-18°) start.
    NauticalTwilightStarts : AstronomyEventCode
        Nautical twilight (-12°) start.
    CivilTwilightStarts : AstronomyEventCode
        Civil twilight (-6°) start.
    Rise : AstronomyEventCode
        Astronomy object rising.
    Meridian : AstronomyEventCode
        Meridian passing (noon).
    AntiMeridian : AstronomyEventCode
        Antimeridian passing (midnight).
    Set : AstronomyEventCode
        Astronomy object setting.
    CivilTwilightEnds : AstronomyEventCode
        Civil twilight (-6°) end.
    NauticalTwilightEnds : AstronomyEventCode
        Nautical twilight (-12°) end.
    AstronomicalTwilightEnds : AstronomyEventCode
        Astronomical twilight (-18°) end.
    NewMoon : AstronomyEventCode
        New moon.
    FirstQuarter : AstronomyEventCode
        Moon in first quarter.
    FullMoon : AstronomyEventCode
        Full moon.
    ThirdQuarter : AstronomyEventCode
        Moon in third quarter.
    """

    AstronomicalTwilightStarts = 1
    NauticalTwilightStarts = 1 << 1
    CivilTwilightStarts = 1 << 2
    Rise = 1 << 3
    Meridian = 1 << 4
    AntiMeridian = 1 << 5
    Set = 1 << 6
    CivilTwilightEnds = 1 << 7
    NauticalTwilightEnds = 1 << 8
    AstronomicalTwilightEnds = 1 << 9
    NewMoon = 1 << 10
    FirstQuarter = 1 << 11
    FullMoon = 1 << 12
    ThirdQuarter = 1 << 13

    @staticmethod
    def resolve(event_code: str) -> "AstronomyEventCode":
        if event_code == "twi18_start":
            return AstronomyEventCode.AstronomicalTwilightStarts
        elif event_code == "twi12_start":
            return AstronomyEventCode.NauticalTwilightStarts
        elif event_code == "twi6_start":
            return AstronomyEventCode.CivilTwilightStarts
        elif event_code == "rise":
            return AstronomyEventCode.Rise
        elif event_code == "meridian":
            return AstronomyEventCode.Meridian
        elif event_code == "antimeridian":
            return AstronomyEventCode.AntiMeridian
        elif event_code == "set":
            return AstronomyEventCode.Set
        elif event_code == "twi6_end":
            return AstronomyEventCode.CivilTwilightEnds
        elif event_code == "twi12_end":
            return AstronomyEventCode.NauticalTwilightEnds
        elif event_code == "twi18_end":
            return AstronomyEventCode.AstronomicalTwilightEnds
        elif event_code == "newmoon":
            return AstronomyEventCode.NewMoon
        elif event_code == "firstquarter":
            return AstronomyEventCode.FirstQuarter
        elif event_code == "fullmoon":
            return AstronomyEventCode.FullMoon
        elif event_code == "thirdquarter":
            return AstronomyEventCode.ThirdQuarter
        else:
            raise ValueError("event_code does not conform to enum AstronomyEventCode")




from libtad.base_service import BaseService
from libtad.datatypes.time import TADDateTime
from libtad.datatypes.astro import AstronomyLocation, AstronomyObjectType
from libtad.datatypes.places import LocationId
from libtad.common import XmlUtils
import libtad.constants as Constants
from typing import List, Dict, Union
import xml.etree.ElementTree as ET
from urllib.parse import urlencode
from urllib.request import urlopen, Request
from ssl import SSLContext

class AstrodataService(BaseService):
    """
    The astrodata service can be used retrieve the altitude, azimuth and 
    distance to the Moon and the Sun for all locations in our database. The 
    service also returns the moon phase, the fraction of the Moons illuminated 
    side as well as the midpoint angle of the Moons bright limb at any point in 
    time.

    ...

    Attributes
    ----------
    is_localtime : bool
        Whether or not the intervals is in local time for the locations queried 
        or UTC time.        
    include_isotime : bool
        Adds time stamps (local time) in ISO 8601 format to all events.
    include_utctime : bool
        Adds UTC time stamps in ISO 8601 format to all events.
    radius : int
        Search radius for translating coordinates (parameter placeid) to locations.
        Coordinates that could not be translated will yield results for the actual
        geographical position – if you would like to query for times at an exact
        location, specify a radius of zero (0).

        The radius in kilometers. Default is infinite, but only locations within
        the same country and time zone are considered.
    language : list of str
        The preferred language(s) for the texts. An error will be raised if the
        language code cannot be recognized. In case the text for a specific event
        cannot be retrieved in any of the requested languages it will be returned
        in English instead.

        It is possible to query all available languages in a single request by
        specifying a primary text language, followed by the special value _all_.
        This value can not be specified as the first language.

        Example:
        ``service.language = ["de"]``

        Example:
        ``service.language.append("de")``

    Methods
    -------
    get_astrodata(object_type, place_id, interval)
        Gets astronomical data for an object at a specific place on specific 
        points in time.
    """

    def __init__(self, access_key: str, secret_key: str):
        """
        Parameters
        ----------
        access_key : str
            Access key.
        secret_key : str
            Secret key.
        """

        super().__init__(access_key, secret_key, "astrodata")
        self.is_localtime: bool = False
        self.include_isotime: bool = False
        self.include_utctime: bool = False
        self.radius: int = None

    def get_astrodata(self, object_type: AstronomyObjectType, place_id: LocationId, interval: Union[TADDateTime, List[TADDateTime]]) -> List[AstronomyLocation]:
        """
        Gets astronomical data for an object at a specific place on specific 
        points in time.

        Parameters
        ----------
        object_type : AstronomyObjectType
            The astronomical object type.
        place_id : LocationId
            Place identifier.
        interval : TADDateTime or list of TADDateTime
            Points in time to query for, using comma to separate multiple 
            timestamps, in ISO 8601 timestamp format.

        Returns
        -------
        astronomy_locations : list of AstronomyLocation
            A list of astronomical information

        """

        if type(interval) is TADDateTime:
            interval = [interval]

        if type(place_id) is not LocationId or not object_type \
                or not interval or type(interval) is not list or type(interval[0]) is not TADDateTime:
            raise ValueError("An argument is invalid")

        ID = str(place_id)
        if not ID:
            raise ValueError("An argument is invalid")

        args: Dict[str, object] = self._authentication_options.copy()
        args["placeid"] = place_id
        args["object"] = object_type.name.lower()
        args["interval"] = ",".join([i._get_second_precision_str() for i in interval])

        return self.__retrieve_astrodata(args)

    def __retrieve_astrodata(self, args: Dict[str, object]) -> List[AstronomyLocation]:
        arguments: Dict[str, object] = self.__get_optional_arguments(args)
        url: str = Constants.ENTRYPOINT + "/" + self._service_name + "?" + urlencode(arguments)
        req = Request(
            url,
            headers = { "User-Agent": "libtad-py"}
            )

        with urlopen(req, context=SSLContext()) as f:
            result: str = f.read().decode("utf-8")

        return self.__from_xml(result)

    def __get_optional_arguments(self, args: Dict[str, object]) -> Dict[str, object]:
        optional_args: Dict[str, str] = args
        
        optional_args["localtime"] = int(self.is_localtime)
        optional_args["isotime"] = int(self.include_isotime)
        optional_args["lang"] = ",".join(self.language)
        optional_args["utctime"] = int(self.include_utctime)
        optional_args["out"] = Constants.DEFAULTRETURNFORMAT
        optional_args["verbosetime"] = str(Constants.DEFAULTVERBOSETIMEVALUE)
        optional_args["version"] = str(self._version)

        if self.radius is not None:
            optional_args["radius"] = str(self.radius)

        return optional_args

    def __from_xml(self, result: str) -> List[AstronomyLocation]:
        XmlUtils.check_for_errors(result)
        xml: ET.Element = ET.fromstring(result)
        return [AstronomyLocation(location) for location in xml.findall("location")]




from enum import Flag

class AstronomyObjectType(Flag):
    """
    An enum class with simple bitflags so the user can easily put together
    a set of types which should be included or excluded in the retrieval 
    of holidays.

    ...

    Attributes
    ----------
    Sun : AstronomyObjectType
        The sun.        
    Moon : AstronomyObjectType
        The moon.
    Mercury : AstronomyObjectType
        Mercury.
    Venus : AstronomyObjectType
        Venus.
    Mars : AstronomyObjectType
        Mars.
    Jupiter : AstronomyObjectType
        Jupiter.
    Saturn : AstronomyObjectType
        Saturn.
    Uranus : AstronomyObjectType
        Uranus.
    Neptune : AstronomyObjectType
        Neptune.
    Pluto : AstronomyObjectType
        Pluto.
    """

    Sun = 1
    Moon = 1 << 1
    Mercury = 1 << 2
    Venus = 1 << 3
    Mars = 1 << 4
    Jupiter = 1 << 5
    Saturn = 1 << 6
    Uranus = 1 << 7
    Neptune = 1 << 8
    Pluto = 1 << 9




from libtad.common.exceptions import ServerSideException, MalformedXMLException
import xml.etree.ElementTree as ET

class XmlUtils:
    @staticmethod
    def check_for_errors(result: str) -> None:
        node: ET.Element = ET.fromstring(result)

        if not node.get("version"):
            raise MalformedXMLException("Expected 'version' attribute in data node")

        error_node: ET.Element = node.find("error")
        if error_node is not None:
            XmlUtils.handle_error(error_node)

    @staticmethod
    def handle_error(error_node: ET.Element) -> None:
        error_msg: str = error_node.text if error_node is not None else "Unspecified error"
        raise ServerSideException(error_msg)
        



MICROINSEC = 100000
SEC = 1
SECINMIN = 60
SECINHOUR = SECINMIN * 60
SECINDAY = SECINHOUR * 24


class TADTimeSpan:
    '''
    A class used to store timespans.
    
    ...

    Methods
    -------
    get_in_microseconds()
        Used to get timespan in microseconds.
    get_in_seconds()
        Used to get timespan in seconds.
    get_in_minutes()
        Used to get timespan in minutes.
    get_in_hours()
        Used to get timespan in hours.
    get_in_days()
        Used to get timespan in days.
    add(days, hours, minutes, seconds, microseconds)
        Adds a given timespan.
    '''

    def __init__(self,
                 days: int = 0,
                 hours: int = 0,
                 minutes: int = 0,
                 seconds: int = 0,
                 microseconds: int = 0):
        '''
        Parameters
        ----------
        days : str
            Days.
        hours : str
            Hours.
        minutes : str
            Minutes.
        seconds : str
            Seconds.
        microseconds : str
            Microseconds.
        '''

        self.__seconds: int = seconds
        self.__seconds += minutes * SECINMIN
        self.__seconds += hours * SECINHOUR
        self.__seconds += days * SECINDAY
        self.__microseconds: int = microseconds
    
    def get_in_microseconds(self) -> int:
        return self.__microseconds + self.__seconds * MICROINSEC

    def get_in_seconds(self) -> int:
        return self.__seconds

    def get_in_minutes(self) -> int:
        return self.__seconds // SECINMIN

    def get_in_hours(self) -> int:
        return self.__seconds // SECINHOUR

    def get_in_days(self) -> int:
        return self.__seconds // SECINDAY

    def add(self,
            days: int = 0,
            hours: int = 0,
            minutes: int = 0,
            seconds: int = 0,
            microseconds: int = 0) -> None: 

        self.__seconds += seconds
        self.__seconds += minutes * SECINMIN
        self.__seconds += hours * SECINHOUR
        self.__seconds += days * SECINDAY
        self.__microseconds += microseconds

    def _parse_and_add(self, offset: str) -> None:
        if not offset:
            raise ValueError("Cannot parse TimeSpan string")

        if "-" in offset:
            offset_prefix = -1
            offset = offset[1:]
        else:
            if "+" in offset:
                offset = offset[1:]
            offset_prefix = 1

        try:
            offset_list = [int(num) * offset_prefix for num in offset.split(":")]
            if len(offset_list) == 2:
                offset_list.append(0)
            if len(offset_list) != 3:
                raise ValueError
        except ValueError:
            raise ValueError("Cannot parse TimeSpan string")

        self.add(hours=offset_list[0], minutes=offset_list[1], seconds=offset_list[2])

    @staticmethod
    def _parse(offset: str) -> "TADTimeSpan":
        time_span = TADTimeSpan()
        time_span._parse_and_add(offset)
        return time_span




from typing import Tuple, Union
from enum import Flag

class AstronomyEventClass(Flag):
    """
    An enum class containing astronomical events

    ...

    Attributes
    ----------
    All : AstronomyEventClass
        Combination of all known classes.
    DayLength : AstronomyEventClass
        Day length.
    Meridian : AstronomyEventClass
        Meridian (Noon, highest point) and Anti-Meridian (lowest point) events.
    Phase : AstronomyEventClass
        Moon phase events. Additionally to the phase events (only occurring on
        four days per lunar month), an additional attribute for the current moon
        phase is reported for every day.
    SetRise : AstronomyEventClass
        Set and rise events. Event times take atmospheric refraction into account.   
    AllTwilights : AstronomyEventClass
        Combination of all 3 twilight classes.        
    CivilTwilight : AstronomyEventClass
        Civil twilight (-6°).
    NauticalTwilight : AstronomyEventClass
        Nautical twilight (-12°).
    AstronomicalTwilight : AstronomyEventClass
        Astronomical twilight (-18°).
    Current : AstronomyEventClass
        The current phase for the place requested. Additional attributes for 
        illumination (moon), azimuth, distance. 
        
    """

    All = 1
    DayLength = 1 << 1
    Meridian = 1 << 2
    Phase = 1 << 3
    SetRise = 1 << 4
    AllTwilights = 1 << 5
    CivilTwilight = 1 << 6
    NauticalTwilight = 1 << 7
    AstronomicalTwilight = 1 << 8
    Current = 1 << 9

    @staticmethod
    def resolve(astro_type: Union[int, str, "AstronomyEventClass"]) -> Tuple[int, str, "AstronomyEventClass"]:
        tup: Tuple[Tuple[int, str, AstronomyEventClass], ...] = (
            (1, "all", AstronomyEventClass.All),
            (1 << 1, "daylength", AstronomyEventClass.DayLength),
            (1 << 2, "meridian", AstronomyEventClass.Meridian),
            (1 << 3, "phase", AstronomyEventClass.Phase),
            (1 << 4, "setrise", AstronomyEventClass.SetRise),
            (1 << 5, "twilight", AstronomyEventClass.AllTwilights),
            (1 << 6, "twilight6", AstronomyEventClass.CivilTwilight),
            (1 << 7, "twilight12", AstronomyEventClass.NauticalTwilight),
            (1 << 8, "twilight18", AstronomyEventClass.AstronomicalTwilight),
            (1 << 9, "current", AstronomyEventClass.Current)
            )

        if type(astro_type) is int:
            index = 0
        elif type(astro_type) is str:
            index = 1
        elif type(astro_type) is AstronomyEventClass:
            index = 2
        else:
            return None

        for astro_entry in tup:
            if astro_type == astro_entry[index]:
                return astro_entry




import libtad.constants as Constants
from libtad.base_service import BaseService
from libtad.datatypes.business import BusinessDaysOperatorType, BusinessDaysFilterType, BusinessDates
from libtad.datatypes.places import LocationId
from libtad.datatypes.time import TADDateTime
from libtad.common import XmlUtils
import xml.etree.ElementTree as ET
from urllib.parse import urlencode
from urllib.request import urlopen, Request
from ssl import SSLContext
from typing import List, Dict, Union

class BusinessDurationService(BaseService):
    """
    The businessduration service can be used to calculate the number of business
    days between a specified start date and end date.

    When you query the businessduration service with a placeid or a country, a 
    start date and an end date the service will return the number of business 
    days in that date range by excluding public holidays and weekends. 
    Furthermore, you can apply additional filters such as individual days and 
    whether or not the calculation should include the filter result or 
    exclude it.

    ...

    Attributes
    ----------
    include : bool
        Specify whether the result should be calculated by including instead
        of excluding the days.
 
    filter : BusinessDaysFilterType or list of BusinessDaysFilterType
        Choose a set of types or days you want to filter on.

    include_last_date : bool
        Whether or not the last date (end_date) should be counted in the result.

    language : list of str
        The preferred language(s) for the texts. An error will be raised if the
        language code cannot be recognized. In case the text for a specific event
        cannot be retrieved in any of the requested languages it will be returned
        in English instead.

        It is possible to query all available languages in a single request by
        specifying a primary text language, followed by the special value _all_.
        This value can not be specified as the first language.

        Example:
        ``service.language = ["de"]``

        Example:
        ``service.language.append("de")``

    Methods
    -------
    get_business_duration_for_place(place_id, start_date, days)
        Gets the business dates by place id.
    get_business_duration_for_country(country_iso, start_date, days, state_iso=None)
        Gets the business dates by country and an optional state.
    """

    def __init__(self, access_key: str, secret_key: str):
        """
        Parameters
        ----------
        access_key : str
            Access key.
        secret_key : str
            Secret key.
        """

        super().__init__(access_key, secret_key, "businessduration")
        self.include: bool = False
        self.filter: Union[BusinessDaysFilterType, List[BusinessDaysFilterType]] = BusinessDaysFilterType.Weekendholidays
        self.include_last_date: bool = False

    def get_business_duration_for_place(self, place_id: LocationId, start_date: TADDateTime, end_date: TADDateTime) -> BusinessDates:
        """
        Gets the business dates by place id.

        Parameters
        ----------
        place_id : LocationId
            Place identifier.
        start_date : TADDateTime
            The first date to count.
        end_date : TADDateTime
            The last date to count.

        Returns
        -------
        business_dates : BusinessDates
            Object containing geographical information and a list of the
            calculated result for the requested period.

        """

        if not isinstance(place_id, LocationId):
            raise ValueError("An argument is invalid")

        ID = str(place_id)
        if not ID:
            raise ValueError("An argument is invalid")

        args: Dict[str, object] = {"placeid": ID}
        return self.__get_business_duration(args, start_date, end_date)

    def get_business_duration_for_country(self, country_iso: str, start_date: TADDateTime, end_date: TADDateTime, state_iso: str = None) -> BusinessDates:
        """
        Gets the business dates by country and an optional state.

        Parameters
        ----------
        country_iso : LocationId
            The country for which you would like to calculate the business date.
        start_date : TADDateTime
            The first date to count.
        end_date : TADDateTime
            The last date to count.
        state_iso : str, optional
            The state in the given country you want to calculate the business date.
            Uses no states as default.

        Returns
        -------
        business_dates : BusinessDates
            Object containing geographical information and a list of the
            calculated result for the requested period.

        """

        if not (isinstance(country_iso, str) and (not state_iso or isinstance(state_iso, str))):
            raise ValueError("An argument is invalid")

        args: Dict[str, object] = {"country": country_iso}
        if (state_iso):
            args["state"] = state_iso

        return self.__get_business_duration(args, start_date, end_date)

    def __get_business_duration(self, args: Dict[str, object], start_date: TADDateTime, end_date: TADDateTime) -> BusinessDates:
        if not isinstance(start_date, TADDateTime) or not isinstance(end_date, TADDateTime):
            raise ValueError("An argument is invalid")

        if end_date < start_date:
            raise ValueError("End date cannot be before start date")

        args.update(self._authentication_options)
        args["startdt"] = str(start_date)
        args["enddt"] = str(end_date)

        return self.__retrieve_business_duration(args)

    def __retrieve_business_duration(self, args: Dict[str, object]) -> BusinessDates:
        arguments: Dict[str, object] = self.__get_optional_arguments(args)
        url: str = Constants.ENTRYPOINT + "/" + self._service_name + "?" + urlencode(arguments)
        req = Request(
                url,
                headers = {"User-Agent": "libtad-py"}
                )

        with urlopen(req, context=SSLContext()) as f:
            result: str = f.read().decode("utf-8")

        return self.__from_xml(result)

    def __get_optional_arguments(self, args: Dict[str, object]) -> Dict[str, object]:
        optional_args: Dict[str, str] = args

        optional_args["include"] = int(self.include)
        optional_args["includelastdate"] = int(self.include_last_date)
        optional_args["lang"] = ",".join(self.language)
        optional_args["out"] = Constants.DEFAULTRETURNFORMAT
        optional_args["verbosetime"] = str(Constants.DEFAULTVERBOSETIMEVALUE)
        optional_args["version"] = str(self._version)

        if isinstance(self.filter, BusinessDaysFilterType):
            optional_args["filter"] = str(self.filter)
        elif isinstance(self.filter, list) and all(isinstance(filter, BusinessDaysFilterType) for filter in self.filter):
            optional_args["filter"] = ",".join(map(str, self.filter))

        return optional_args

    def __from_xml(self, result: str) -> BusinessDates:
        XmlUtils.check_for_errors(result)
        xml: ET.Element = ET.fromstring(result)
        return BusinessDates(xml.find("geo"), [period for period in xml.findall("period")])




from .geo import Geo
from libtad.datatypes.astro import Astronomy
from libtad.datatypes.time import TimeChange, TADTime
import xml.etree.ElementTree as ET
from typing import List

class Location:
    """
    An object containing data for location.

    ...

    Attributes
    ----------
    id : str
        The id of the location.
    geography : Geo
        Geographical information about the location.
    time : TADTime
        Time information about the location. Only present if requested.
    time_changes : list of TimeChange
        Time changes (daylight savings time). Only present if requested and 
        information is available.
    astronomy : list of Astronomy
        Astronomical information – sunrise and sunset times.
        Only for the timeservice and if requested.
    """

    def __init__(self, node: ET.Element):
        self.id: str = None
        self.geography: Geo = None
        self.time: TADTime = None
        self.time_changes: List[TimeChange] = []
        self.astronomy: List[Astronomy] = []

        ID = node.get("id")
        geo = node.find("geo")
        time = node.find("time")
        timechanges = node.find("timechanges")
        astronomy = node.find("astronomy")

        if ID:
            self.id = ID

        if geo is not None:
            self.geography = Geo(geo)

        if time is not None:
            self.time = TADTime(time)

        if timechanges is not None:
            self.time_changes = [TimeChange(change) for change in timechanges.findall("change")]

        if astronomy is not None:
            self.astronomy = [Astronomy(obj) for obj in astronomy.findall("object")]




from .astronomy_special_type import AstronomySpecialType
from libtad.common.exceptions import MalformedXMLException
import xml.etree.ElementTree as ET

class AstronomySpecial:
    """
    This element is only present if there are no astronomical events.

    ...

    Attributes
    ----------
    type : AstronomySpecialType
        Indicates if the sun is up or down all day.
    """

    def __init__(self, node: ET.Element):
        self.type: AstronomySpecialType = None

        type_ = node.get("type")

        if type_:
            type_cap = type_.capitalize()
            try:
                self.type = AstronomySpecialType[type_.capitalize()]
            except KeyError:
                pass

        if not self.type:
            raise MalformedXMLException("The XML Received from Time and Date did not include an event type which complies with an AstronomySpecialType enum", 
                    ignore_default=True)




from .astronomy_object_type import AstronomyObjectType
from .astronomy_day import AstronomyDay
from .astronomy_current import AstronomyCurrent
from libtad.common.exceptions import MalformedXMLException
import xml.etree.ElementTree as ET
from typing import List

class AstronomyObjectDetails:
    """
    A class used to store astronomy objects

    ...

    Attributes
    ----------
    name : AstronomyObjectType
        Object name.
    days : list of AstronomyDay
        astronomy service: Lists all the requested days where events are happening.

        astrodata service: N/A
    current : AstronomyCurrent
        astronomy service: The current data for the object. Only if requested.

        astrodata service: N/A
    current : list of AstronomyCurrent
        astronomy service: N/A

        astrodata service: The specific data for the object at isotime/utctime.

    """

    def __init__(self, node: ET.Element):
        self.name: AstronomyObjectType = AstronomyObjectType(0)
        self.days: List[AstronomyDay] = None
        self.current: AstronomyCurrent = None
        self.result: List[AstronomyCurrent] = None

        name = node.get("name")
        days = node.findall("day")
        current = node.find("current")
        results = node.findall("result")
        
        if name and name.capitalize() in AstronomyObjectType.__members__:
            self.name = AstronomyObjectType[name.capitalize()]
        else:
            raise MalformedXMLException(name)
        
        if days:
            self.days = [AstronomyDay(day) for day in days]

        if current is not None:
            self.current = AstronomyCurrent(current)

        if results:
            self.result = [AstronomyCurrent(result) for result in results]




import xml.etree.ElementTree as ET

class LocationRef:
    """
    A class used for locations referenced by a region.

    ...

    Attributes
    ----------
    id : str
        The id of the location.
    name : str
        The name of the location.
    state : str
        The state of the location within the country (only if applicable).
    """

    def __init__(self, node: ET.Element):
        self.id: str = None
        self.name: str = None
        self.state: str = None

        ID = node.get("id")
        name = node.get("name")
        state = node.get("state")

        if ID:
            self.id = ID

        if name:
            self.name = name

        if state:
            self.state = state




__all__ = [
    "Country",
    "Geo",
    "Place",
    "LocationRef",
    "LocationId",
    "Location",
    "Region",
    "Coordinates",
]

from libtad.datatypes.places.country import Country
from libtad.datatypes.places.geo import Geo
from libtad.datatypes.places.place import Place
from libtad.datatypes.places.location_ref import LocationRef
from libtad.datatypes.places.location_id import LocationId
from libtad.datatypes.places.location import Location
from libtad.datatypes.places.region import Region
from libtad.datatypes.places.coordinates import Coordinates

def __dir__():
    return __all__


from .dst_special_type import DSTSpecialType
from libtad.datatypes.places import Region
from libtad.datatypes.time import TimeChange, TADDateTime, TADTimezone
import xml.etree.ElementTree as ET
from typing import List

class DST:
    """
    DST information for a given region.

    ...

    Attributes
    ----------
    region : Region
        The geographical region where this information is valid.
        Contains country, a textual description of the region and
        the name of the biggest place.
    standard_timezone : TADTimezone
        Information about the standard timezone. This element is
        always returned.
    dst_timezone : TADTimezone
        Information about the daylight savings timezone. Suppressed,
        if there are no DST changes in the queried year.

        Please note that if the region is on daylight savings time
        for the whole year, this information will be returned in the
        stdtimezone element. Additionally, the Special element will
        be set to DaylightSavingTimeAllYear.
    special : DSTSpecialType
        Indicates if the region does not observe DST at all, or is on
        DST all year long.
    dst_start : TADDateTime
        Starting date of daylight savings time. Suppressed, if there
        are no DST changes in the queried year.
    dst_end : TADDateTime
        Ending date of daylight savings time. Suppressed, if there are
        no DST changes in the queried year.
    time_changes : list of TimeChange
        Time changes (daylight savings time). Only present if requested
        and information is available.
    """

    def __init__(self, node: ET.Element):
        self.region: Region = None
        self.standard_timezone: TADTimezone = None
        self.dst_timezone: TADTimezone = None
        self.special: DSTSpecialType = None
        self.dst_start: TADDateTime = None
        self.dst_end: TADDateTime = None
        self.time_changes: List[TimeChange] = []

        region = node.find("region")
        stdtimezone = node.find("stdtimezone")
        dsttimezone = node.find("dsttimezone")
        special_node = node.find("special")
        special = special_node.get("type") if special_node is not None else None
        dststart = node.find("dststart")
        dstend = node.find("dstend")
        timechanges = node.find("timechanges")

        if region is not None:
            self.region = Region(region)

        if stdtimezone is not None:
            self.standard_timezone = TADTimezone(stdtimezone)
            
        if dsttimezone is not None:
            self.dst_timezone = TADTimezone(dsttimezone)

        if special:
            self.special = DSTSpecialType.resolve(special)

        if dststart is not None:
            self.dst_start = TADDateTime._parse(dststart.text)

        if dstend is not None:
            self.dst_end = TADDateTime._parse(dstend.text)

        if timechanges is not None:
            self.time_changes = [TimeChange(change) for change in timechanges.findall("change")]




from .holiday_state import HolidayState
from libtad.datatypes.time import TADTime
from libtad.datatypes.places import Country
import xml.etree.ElementTree as ET
from urllib.parse import urlparse, ParseResult
from typing import List, Dict

class Holiday:
    '''
    A class used for holidays.
    
    ...

    Attributes
    ----------
    id : int
        Identifier for the holiday definition. Please note that this id
        is not unique, not even with a single year – the same holiday
        event may be returned multiple time because it is observed on a
        different day, or because it is scheduled in a different calendar
        (Hebrew or Muslim calendar) and happens multiple times within a
        Gregorian year. Use the Uid field for purposes where you need a
        unique identifier.
    uid : str
        Id for the shown holiday instance. The id is designed to be unique
        across all holiday instances generated by the libtad.com API
        services and respects different calendars and other reasons that
        may cause events to occurs multiple times within one Gregorian year.

        Example: 0007d600000007db
    name : str
        Holiday/Observance name.
    date : TADTime
        Date/time of the holiday instance. Most holidays do have a specific
        time – in this case the time components will be skipped. Some
        special events like equinoxes and solstices include the exact time
        of the event as well, in this case the timestamp will be in local
        time zone (including time zone data) (countries with multiple
        timezones: local time in capital city).            
    url : ParseResult
        Further information about the specific holiday. The URL points to
        the libtad.com web page.

        Example: http://www.timeanddate.com/holidays/us/new-year-day
    country : Country
        Country of the holiday instance.
    locations : str
        Summary of locations where this holiday instance is valid. Element is
        only present if the holiday instance does not affect the whole country.
    states : list of HolidayState
        States/subdivisions that are affected by this holiday instance. This
        element is only present if the holiday instance is not valid in the
        whole country.
    oneliner : str
        A short description of the holiday instance.
    types : list of str
        Classification of the holiday. Most days have only one classification,
        bust some have multiple types associated. This happens e.g. in
        conjunction with religious days that also are flag days.

        Example: National Holiday
    '''

    def __init__(self, node: ET.Element):
        self.id: int = -1
        self.uid: str = ""
        self.urlid: str = ""
        self.name: Dict[str] = {}
        self.date: TADTime = None
        self.url: ParseResult = None
        self.country: Country = None
        self.locations: str = ""
        self.states: List[HolidayState] = []
        self.oneliner: Dict[str] = {}
        self.types: List[str] = []

        ID = node.get("id")       
        url = node.get("url")
        urlid = node.get("urlid")
        name = node.find("name")
        locations = node.find("locations")
        uid = node.find("uid")
        country = node.find("country")
        oneliner = node.find("oneliner")
        types = node.find("types")
        states = node.find("states")
        date = node.find("date")

        if ID:
            self.id = int(ID)

        if url:
            self.url = urlparse(url)

        if urlid:
            self.urlid = urlid

        if uid is not None:
            self.uid = uid.text
        else:
            uid = node.get("uid")
            if uid:
                self.uid = uid

        if country is not None:
            self.country = Country(country)

        if oneliner is not None:
            for oneliner_node in oneliner:
                if oneliner_node is None:
                    continue
                oneliner_lang = oneliner_node.get("lang")
                oneliner_text = oneliner_node.text
                if oneliner_lang is not None and oneliner_text is not None:
                    self.oneliner[oneliner_lang] = oneliner_text

        if date is not None:
            self.date = TADTime(date)

        if name is not None:
            for name_node in name:
                if name_node is None:
                    continue
                name_lang = name_node.get("lang")
                name_text = name_node.text
                if name_lang is not None and name_text is not None:
                    self.name[name_lang] = name_text

        if locations is not None:
            self.locations = locations.text

        if types is not None:
            for t in types:
                self.types.append(t.text)

        if states is not None:
            for s in states:
                self.states.append(HolidayState(s))




__all__ = [
        "HolidaysService",
        "AstronomyService", 
        "AstrodataService", 
        "DSTService", 
        "ConvertTimeService", 
        "PlacesService", 
        "TimeService", 
        "BusinessDateService", 
        "BusinessDurationService",
        "datatypes.astro",
        "datatypes.business",
        "datatypes.converted_times",
        "datatypes.dst",
        "datatypes.holidays",
        "datatypes.places",
        "datatypes.time",
        ]

from .datatypes import *
from . import common
from . import constants

from libtad.holidays_service import HolidaysService
from libtad.astrodata_service import AstrodataService
from libtad.astronomy_service import AstronomyService
from libtad.dst_service import DSTService
from libtad.convert_time_service import ConvertTimeService
from libtad.places_service import PlacesService
from libtad.time_service import TimeService
from libtad.business_date_service import BusinessDateService
from libtad.business_duration_service import BusinessDurationService

def __dir__():
        return __all__



from libtad.datatypes.places import Geo
from .period_type import PeriodType
import xml.etree.ElementTree as ET
from typing import List

class BusinessDates:
    """
    A class used to store geographical information and a list of 
    calculated results for a period.

    ...

    Attributes
    ----------
    geo : Geo
        Geographical information about the location
    periods : list of PeriodType
        List of calculated results for a period.
    """

    def __init__(self, node_geo: ET.Element, node_periods: List[ET.Element]):
        self.geo = Geo(node_geo)
        self.periods = [PeriodType(period) for period in node_periods]




from .weekdays_type import WeekdaysType
from .business_holiday_type import BusinessHolidayType
from libtad.datatypes.time import TADTime
import xml.etree.ElementTree as ET

class PeriodType:
    """
    Calculated result for a given period.

    ...

    Attributes
    ----------
    included_days : int
        Number of days in the given period.
    calendar_days : int
        Number of calendar days in the given period.
    skipped_days : int
        Number of days which was skipped in the given period.
    start_date : TADTime
        The date the period started from.
    end_date : TADTime
        The date the period ended on.
    weekdays : WeekdaysType
        The spread of excluded or included weekdays in included_days.
    holidays : BusinessHolidayType
        Holidays which occur in the given period.
    """

    def __init__(self, node: ET.Element):
        self.included_days: int = None
        self.calendar_days: int = None
        self.skipped_days: int = None
        self.start_date: TADTime = None
        self.end_date: TADTime = None
        self.weekdays: WeekdaysType = None
        self.holidays: BusinessHolidayType = None

        includeddays = node.get("includeddays")
        calendardays = node.get("calendardays")
        skippeddays = node.get("skippeddays")
        startdate = node.find("startdate")
        enddate = node.find("enddate")
        weekdays = node.find("weekdays")
        holidays = node.find("holidays")

        if includeddays:
            try:
                self.included_days = int(includeddays)
            except ValueError:
                pass

        if calendardays:
            try:
                self.calendar_days = int(calendardays)
            except ValueError:
                pass

        if skippeddays:
            try:
                self.skipped_days = int(skippeddays)
            except ValueError:
                pass

        if startdate is not None:
            self.start_date = TADTime(startdate)

        if enddate is not None:
            self.end_date = TADTime(enddate)

        if weekdays is not None:
            self.weekdays = WeekdaysType(weekdays)
            
        if holidays is not None:
            self.holidays = BusinessHolidayType(holidays)




class TADDateTime:
    '''
    A class used to store date and time.

    ...

    Attributes
    ----------
    year : int
        Year.
    month : int
        Month.
    day : int
        Day.
    hour : int
        Hour.
    minute : int
        Minute.
    second : int
        Second.
    microsecond : int
        Microsecond.
    '''

    def __init__(self,
                 year:          int = 0,
                 month:         int = 1,
                 day:           int = 1,
                 hour:          int = 0,
                 minute:        int = 0,
                 second:        int = 0,
                 microsecond:   int = 0):
        '''
        Parameters
        ----------
        year : int
            Year.
        month : int
            Month.
        day : int
            Day.
        hour : int
            Hour.
        minute : int
            Minute.
        second : int
            Second.
        microsecond : int
            Microsecond.
        '''

        self.__year:        int = 0
        self.__month:       int = 1
        self.__day:         int = 1
        self.__hour:        int = 0
        self.__minute:      int = 0
        self.__second:      int = 0
        self.__microsecond: int = 0
        
        self.year:          int = year
        self.month:         int = month
        self.day:           int = day
        self.hour:          int = hour
        self.minute:        int = minute
        self.second:        int = second
        self.microsecond:   int = microsecond
        
    @property
    def year(self):
        return self.__year
    
    @year.setter
    def year(self, value: int):
        self.__year = value

    @property
    def month(self):
        return self.__month
    
    @month.setter
    def month(self, value: int):
        start = 1
        end = 12
        if start <= value <= end:
            self.__month = value
        else:
            raise ValueError(f"Month must be an integer between {start} and {end}")
                
    @property
    def day(self):
        return self.__day
    
    @day.setter
    def day(self, value: int):
        start = 1
        end = 31
        if start <= value <= end:
            self.__day = value
        else:
            raise ValueError(f"Day must be an integer between {start} and {end}")
                
    @property
    def hour(self):
        return self.__hour
    
    @hour.setter
    def hour(self, value: int):
        start = 0
        end = 24
        if start <= value <= end:
            self.__hour = value
        else:
            raise ValueError(f"Hour must be an integer between {start} and {end}")
                
    @property
    def minute(self):
        return self.__minute
    
    @minute.setter
    def minute(self, value: int):
        start = 0
        end = 59
        if start <= value <= end:
            self.__minute = value
        else:
            raise ValueError(f"Minute must be an integer between {start} and {end}")

    @property
    def second(self):
        return self.__second
    
    @second.setter
    def second(self, value: int):
        start = 0
        end = 60
        if start <= value <= end:
            self.__second = value
        else:
            raise ValueError(f"Second must be an integer between {start} and {end}")

    @property
    def microsecond(self):
        return self.__microsecond
    
    @microsecond.setter
    def microsecond(self, value: int):
        start = 0
        end = 99999
        if start <= value <= end:
            self.__microsecond = value
        else:
            raise ValueError(f"Microsecond must be an integer between {start} and {end}")

    @staticmethod
    def _parse(fmtstr: str) -> "TADDateTime":
        include_time = "T" in fmtstr

        if include_time:
            strlist = fmtstr.split("T")
            if len(strlist) != 2:
                raise ValueError("Cannot parse DateTime string")
            for char in "+-Z":
                strlist[1] = strlist[1].split(char)[0]
        else:
            strlist = [fmtstr]

        try:
            date_list = [int(num) for num in strlist[0].split("-")]
            time_list = [int(num) for num in strlist[1].split(":")] if include_time else [0] * 3
        except ValueError:
            raise ValueError("Cannot parse DateTime string")

        if len(date_list) != 3 or len(time_list) != 3:
            raise ValueError("Cannot parse DateTime string")

        datetime = TADDateTime(year=date_list[0], month=date_list[1], day=date_list[2], hour=time_list[0], minute=time_list[1], second=time_list[2])
        return datetime

    def _get_day_precision_str(self):
        return f"{str(self.__year).zfill(4)}-{str(self.__month).zfill(2)}-{str(self.__day).zfill(2)}"

    def _get_second_precision_str(self):
        return f"{self._get_day_precision_str()}T{str(self.__hour).zfill(2)}:{str(self.__minute).zfill(2)}:{str(self.__second).zfill(2)}"

    def __eq__(self, value):
        if self.__year == value.__year \
                and self.__month == value.month \
                and self.__day == value.day \
                and self.__hour == value.hour \
                and self.__day == value.day \
                and self.__second == value.second \
                and self.__microsecond == value.microsecond:
            return True
        return False

    def __ne__(self, value):
        return not self.__eq__(value)

    def __lt__(self, value):
        for t in ("year", "month", "day", "hour", "day", "second", "microsecond"):
            self_t = getattr(self, t)
            value_t = getattr(value, t)
            if self_t < value_t:
                return True
            elif self_t > value_t:
                return False
        return False 
    
    def __le__(self, value):
        return self.__lt__(value) or self.__eq__(value)

    def __gt__(self, value):
        return not self.__le__(value)

    def __ge__(self, value):
        return not self.__lt__(value)

    def __str__(self):
        return self._get_second_precision_str()




from enum import Flag

class MoonPhase(Flag):
    """
    An enum class containing moon phase for the day.

    ...

    Attributes
    ----------
    NotRequested : MoonPhase
        The not requested moon phases.
    NewMoon : MoonPhase
        New moon.
    WaxingCrescent : MoonPhase
        Waxing crescent moon.
    FirstQuarter : MoonPhase
        Moon in first quarter.
    WaxingGibbous : MoonPhase
        Waxing gibbous moon.
    FullMoon : MoonPhase
        Full moon.
    WaningGibbous : MoonPhase
        Waning gibbous moon.
    ThirdQuarter : MoonPhase
        Moon in third quarter.
    WaningCrescent : MoonPhase
        Waning crescent moon.
    """

    NotRequested = 1
    NewMoon = 1 << 1
    WaxingCrescent = 1 << 2
    FirstQuarter = 1 << 3
    WaxingGibbous = 1 << 4
    FullMoon = 1 << 5
    WaningGibbous = 1 << 6
    ThirdQuarter = 1 << 7
    WaningCrescent = 1 << 8

    @staticmethod
    def _parse(moonphase: str) -> "MoonPhase":
        for phase in MoonPhase:
            if phase.name.lower() != moonphase.lower():
                continue
            return phase




from .country import Country
from .location_ref import LocationRef
import xml.etree.ElementTree as ET

class Region:
    """
    The geographical region. Contains country, a textual description of 
    the region and the name of the biggest place.

    ...

    Attributes
    ----------
    country : Country
        The country.
    description : str
        Textual description of a region.

        Example: All locations

        Example: most of Newfoundland and Labrador

        Example: some regions of Nunavut Territory; small region of Ontario
    biggest_place : str
        Name of the biggest city within the region.
    locations : list of LocationRef
        A list of all locations referenced by this region. Only returned if 
        requested by specifying the parameter listplaces.
    """

    def __init__(self, node: ET.Element):
        self.country: Country = None
        self.description: str = ""
        self.biggest_place: str = ""
        self.locations: List[LocationRef] = []

        country = node.find("country")
        desc = node.find("desc")
        biggest_place = node.find("biggestplace")
        locations = node.find("locations")

        if country is not None:
            self.country = Country(country)

        if desc is not None:
            self.description = desc.text

        if biggest_place is not None:
            self.biggest_place = biggest_place.text

        if locations is not None:
            self.locations = [LocationRef(location) for location in locations.findall("location")]




from .business_filter_method import BusinessFilterMethod
import xml.etree.ElementTree as ET

class WeekdaysType:
    """
    The spread of excluded or included weekdays.

    ...

    type : BusinessFilterMethod
        Specifies whether or not the weekdays counted were part of an included 
        or excluded filter.
    count : int
        How many days in total have been counted.
    mon : int
        Count for Mondays.
    tue : int
        Count for Tuesdays.
    wed : int
        Count for Wednesdays.
    thu : int
        Count for Thursdays.
    fri : int
        Count for Fridays.
    sat : int
        Count for Saturdays.
    sun : int
        Count for Sundays.
    """

    def __init__(self, node: ET.Element):
        self.type: BusinessFilterMethod = None
        self.count: int = None
        self.mon: int = None
        self.tue: int = None
        self.wed: int = None
        self.thu: int = None
        self.fri: int = None
        self.sat: int = None
        self.sun: int = None

        _type = node.get("type")
        count = node.get("count")
        mon = node.find("mon")
        tue = node.find("tue")
        wed = node.find("wed")
        thu = node.find("thu")
        fri = node.find("fri")
        sat = node.find("sat")
        sun = node.find("sun")

        if _type:
            try:
                self.type = BusinessFilterMethod[_type.capitalize()]
            except KeyError:
                pass

        if count:
            try:
                self.count = int(count)
            except ValueError:
                pass

        if mon is not None:
            try:
                self.mon = int(mon.text)
            except ValueError:
                pass

        if tue is not None:
            try:
                self.tue = int(tue.text)
            except ValueError:
                pass

        if wed is not None:
            try:
                self.wed = int(wed.text)
            except ValueError:
                pass

        if thu is not None:
            try:
                self.thu = int(thu.text)
            except ValueError:
                pass

        if fri is not None:
            try:
                self.fri = int(fri.text)
            except ValueError:
                pass

        if sat is not None:
            try:
                self.sat = int(sat.text)
            except ValueError:
                pass

        if sun is not None:
            try:
                self.sun = int(sun.text)
            except ValueError:
                pass




from .astronomy_object_type import AstronomyObjectType
from .astronomy_event import AstronomyEvent
from .astronomy_event_type import AstronomyEventType
from .astronomy_special import AstronomySpecial
from libtad.datatypes.time import TADDateTime
from libtad.common.exceptions import MalformedXMLException
import xml.etree.ElementTree as ET
from typing import List, Union

class Astronomy:
    """
    Astronomical information – sunrise and sunset times. 
    Only for the timeservice and if requested.

    ...

    Attributes
    ----------
    name : AstronomyObjectType
        Object name. Currently, the sun is the only supported astronomical object.
    events : list of AstronomyEvent
        Lists all sunrise/sunset events during the day.
    special : AstronomySpecial
        This element is only present if there are no astronomical events. In this
        case it will indicate if the sun is up or down the whole day.
    """

    def __init__(self, node: ET.Element):
        self.name: AstronomyObjectType = None
        self.events: List[AstronomyEvent] = []
        self.special: AstronomySpecial = None

        name = node.get("name")
        events = node.findall("event")
        special = node.find("special")

        if name:
            object_name = name.capitalize()
            try:
                self.name = AstronomyObjectType[object_name]
            except KeyError:
                pass

        if not self.name:
            raise MalformedXMLException(f"The XML Received from Time and Date did not include an object name which complies with an AstronomyObjectType enum: {name}", 
                    ignore_default=True)

        if events:
            for event in events:
                self.events.append(AstronomyEvent(event))

        if special is not None:
            self.special = AstronomySpecial(special)
            

    @property
    def sunset(self) -> Union[TADDateTime, None]:
        """
        This returns the hour and minute of the sunset in TADDateTime format. If 
        there is no sunset, None will be returned, but the Special-property will 
        say whether or not the sun is up or down.

        Returns
        -------
        sunset : TADDateTime or None
            The sunset in TADDateTime. None if there is no sunset that day.
        """
        sets = [event for event in self.events if event.type == AstronomyEventType.Set]
        if len(sets) == 1:
            return sets[0].time
        elif len(sets) > 1:
            sets.sort(key=lambda x : x.time)
            return sets[-1].time
        else:
            return None

    @property
    def sunrise(self) -> Union[TADDateTime, None]:
        """
        This returns the hour and minute of the sunrise in TADDateTime format. If 
        there is no sunrise, None will be returned, but the Special-property will 
        say whether or not the sun is up or down.

        Returns
        -------
        sunrise : TADDateTime or None
            The sunrise in TADDateTime. None if there is no sunrise that day.
        """
        rises = [event for event in self.events if event.type == AstronomyEventType.Rise]
        if len(rises) == 1:
            return rises[0].time
        elif len(rises) > 1:
            rises.sort(key=lambda x : x.time)
            return rises[-1].time
        else:
            return None




from enum import Enum

class BusinessFilterMethod(Enum):
    """
    An enum class containing methods for filtering in business.PeriodTypeapis.

    ...

    Attributes
    ----------
    Included : BusinessFilterMethod
        Included.
    Excluded : BusinessFilterMethod
        Excluded.
    """

    Included = 1
    Excluded = 2




import xml.etree.ElementTree as ET

class HolidayState:
    '''
    A class used for holidays in states

    ...

    Attributes
    ----------
    id : int
        Unique id of the state/subdivision. (default is -1)
    iso : str
        An ISO 3166-1 country or ISO 3166-2 country state code (see [ISO3166]).
    abbreviation : str
        Abbreviation of the state/subdivision.
    name : str
        Common name of the state/subdivision.
    exception : str
        Eventual exception if the holiday does not affect
        the whole state/subdivision.
    '''
    def __init__(self, node: ET.Element):
        self.id: int = -1
        self.iso: str = ""
        self.abbreviation: str = ""
        self.name: str = ""
        self.exception: str = ""

        ID = node.find("id")
        iso = node.get("iso")
        abbr = node.find("abbrev")
        name = node.find("name")
        excp = node.find("exception")

        if ID is not None:
            self.id = int(ID.text)

        if iso:
            self.iso = iso

        if abbr is not None:
            self.abbreviation = abbr.text

        if name is not None:
            self.name = name.text

        if excp is not None:
            self.exception = excp.text
    



from .astronomy_event_code import AstronomyEventCode
from libtad.datatypes.time import TADDateTimeOffset
import xml.etree.ElementTree as ET

class AstronomyDayEvent:
    """
    A class used to store events happening at a given day.

    ...

    Attributes
    ----------
    type : AstronomyEventCode
        Indicates the type of the event.
    isotime : TADDateTimeOffset
        Local time at which the event is happening (including UTC offset).
        The time does not include the seconds.
    utctime : TADDateTimeOffset
        UTC time at which the event is happening. The time does not include
        the seconds.
    altitude : float
        Altitude of the center of the queried astronomical object above an
        ideal horizon.

        Only for meridian type events.
    azimuth : float
        Horizontal direction of the astronomical object at set/rise time
        (referring to true north). North is 0 degrees, east is 90 degrees,
        south is 180 degrees and west is 270 degrees.

        Only for rise and set type events.
    distance : float
        Distance of the earth's center to the center of the queried
        astronomical object in kilometers.

        Only for meridian type events.
    illuminated : float
        The fraction of the Moon's surface illuminated by the Sun's
        rays as seen from the selected location.

        Only for the moon for meridian type events.
    posangle : float
        The counterclockwise angle of the midpoint of the Moon's bright limb as seen from the selected location.

        Only for the moon for meridian type events.

    """

    def __init__(self, node: ET.Element):
        self.type: AstronomyEventCode = None
        self.isotime: TADDateTimeOffset = None
        self.utctime: TADDateTimeOffset = None
        self.altitude: float = None
        self.azimuth: float = None
        self.distance: float = None
        self.illuminated: float = None
        self.posangle: float = None
        
        astro_type = node.get("type")
        isotime = node.get("isotime")
        utctime = node.get("utctime")
        altitude = node.get("altitude")
        azimuth = node.get("azimuth")
        distance = node.get("distance")
        illuminated = node.get("illuminated")
        posangle = node.get("posangle")

        if astro_type:
            self.type = AstronomyEventCode.resolve(astro_type)

        if isotime:
            self.isotime = TADDateTimeOffset._parse(isotime)

        if utctime:
            self.utctime = TADDateTimeOffset._parse(utctime)
        
        if altitude:
            self.altitude = float(altitude)

        if azimuth:
            self.azimuth = float(azimuth)

        if distance:
            self.distance = float(distance)

        if illuminated:
            self.illuminated = float(illuminated)

        if posangle:
            self.posangle = float(posangle)




from .astronomy_object_details import AstronomyObjectDetails
from libtad.datatypes.places import Geo
import xml.etree.ElementTree as ET
from typing import List

class AstronomyLocation:
    """
    A class that wraps the information for the locations in the request.

    ...

    Attributes
    ----------
    id : str
        The id of the location.
    geography : Geo
        Geographical information about the location.
    objects : list of AstronomyObjectDetails
        Requested astronomical information.
    """

    def __init__(self, node: ET.Element):
        self.id: str = ""
        self.matchparam: str = ""
        self.geography: Geo = None
        self.objects: List[AstronomyObjectDetails] = []

        ID = node.get("id")
        matchparam = node.get("matchparam")
        geo = node.find("geo")
        astro = node.find("astronomy")

        if ID:
            self.id = ID

        if matchparam:
            self.matchparam = matchparam

        if geo is not None:
            self.geography = Geo(geo)

        if astro is not None:
            for object_node in astro:
                if object_node is not None:
                    self.objects.append(AstronomyObjectDetails(object_node))




from libtad.datatypes.time import TADTime
from libtad.datatypes.places import Location
import xml.etree.ElementTree as ET
from typing import List

class ConvertedTimes:
    """
    An object containing UTC time stamp of requested time and a list of time
    information for the locations in the request.

    ...

    Attributes
    ----------
    utc : TADTime
        UTC time stamp of requested time.
    locations : list of Location
        This element contains the time information for the locations mentioned 
        in the request.
    """

    def __init__(self, node_utc: ET.Element, node_locations: List[ET.Element]):
        self.utc: TADTime = None
        self.locations: List[Location] = []

        time = node_utc.find("time")

        if time is not None:
            self.utc = TADTime(time)

        if node_locations:
            self.locations = [Location(loc) for loc in node_locations]




from enum import Enum

class BusinessDaysOperatorType(Enum):
    """
    An enum class containing business.PeriodTypedays operators.

    ...

    Attributes
    ----------
    Add : BusinessDaysOperatorType
        Add.
    Subtract : BusinessDaysOperatorType
        Subtract.
    """

    Add = 1
    Subtract = 2




__all__ = [
    "TADTime",
    "TADTimezone",
    "TADDateTime",
    "TADDateTimeOffset",
    "TADTimeSpan",
    "TimeChange",
]

from libtad.datatypes.time.tad_time import TADTime
from libtad.datatypes.time.tad_timezone import TADTimezone
from libtad.datatypes.time.tad_date_time import TADDateTime
from libtad.datatypes.time.tad_date_time_offset import TADDateTimeOffset
from libtad.datatypes.time.tad_time_span import TADTimeSpan
from libtad.datatypes.time.time_change import TimeChange

def __dir__():
    return __all__



from libtad.base_service import BaseService
from libtad.datatypes.holidays import HolidayType, Holiday
from libtad.common import XmlUtils
import libtad.constants as Constants
import xml.etree.ElementTree as ET
from urllib.parse import ParseResult, urlunparse, urlencode
from urllib.request import urlopen, Request
from ssl import SSLContext
from datetime import datetime
from typing import Dict, List


class HolidaysService(BaseService):
    """
    The holidays service can be used to retrieve the list of holidays for a country.

    ...

    Attributes
    ----------
    types : HolidayType
        Holiday types which shoud be returned.
        To combine multiple classes, use the binary OR operator.

        Example:
        ``service.IncludedHolidayTypes = HolidayType.Local``
    language : list of str
        The preferred language(s) for the texts. An error will be raised if the
        language code cannot be recognized. In case the text for a specific event
        cannot be retrieved in any of the requested languages it will be returned
        in English instead.

        It is possible to query all available languages in a single request by
        specifying a primary text language, followed by the special value _all_.
        This value can not be specified as the first language.

        Example:
        ``service.language = ["de"]``

        Example:
        ``service.language.append("de")``

    Methods
    -------
    holidays_for_country(country_code, year=datetime.now().year)
        The holidays service can be used to retrieve the list of holidays for a country.
    """

    def __init__(self, access_key: str, secret_key: str):
        """
        Parameters
        ----------
        access_key : str
            Access key.
        secret_key : str
            Secret key.
        """

        super().__init__(access_key, secret_key, "holidays")
        self.types: HolidayType = HolidayType(0)

    def __get_holiday_types(self) -> str:
        included_strings: List[str] = []
        for hol_item in HolidayType:
            if self.types & hol_item:
                included_strings.append(HolidayType.resolve(hol_item)[1])
        included: str = ",".join(included_strings)
        return included

    def __from_xml(self, result: str) -> List[Holiday]:
        XmlUtils.check_for_errors(result)
        xml: ET.Element = ET.fromstring(result)
        return [Holiday(hol_node) for hol_node in xml.find("holidays")]

    def __get_arguments(self, country_code: str, year: int) -> Dict[str, str]:
        args: Dict[str, object] = self._authentication_options.copy()
        types: str = self.__get_holiday_types()
        args["country"] = country_code
        args["lang"] = ",".join(self.language)
        args["version"] = str(self._version)
        args["verbosetime"] = str(Constants.DEFAULTVERBOSETIMEVALUE)
        args["out"] = Constants.DEFAULTRETURNFORMAT

        if types:
            args["types"] = types

        if year > 0:
            args["year"] = str(year)

        return args

    def __retrieve_holidays(self, country_code: str, year: int) -> List[Holiday]:
        arguments: Dict[str, str] = self.__get_arguments(country_code, year)
        url: str = Constants.ENTRYPOINT + "/" + self._service_name + "?" + urlencode(arguments)
        req = Request(
                url,
                headers = { "User-Agent": "libtad-py"}
                )

        with urlopen(req, context=SSLContext()) as f:
            result: str = f.read().decode("utf-8")

        return self.__from_xml(result)

    def holidays_for_country(self, country_code: str, year: int = datetime.now().year) -> List[Holiday]:
        """
        The holidays service can be used to retrieve the list of holidays for a country.
        If the argument `year` is not passed in, the current year is used.

        Parameters
        ----------
        country_code : str 
            Specify the ISO3166-1-alpha-2 Country Code for which you would like to 
            retrieve the list of holidays.
        year : int, optional
            The year for which the holidays should be retrieved.            
            Uses the current year by default.

        Returns
        -------
        holiday_list : list of Holiday
            List of holidays for a given country.

        """

        if not country_code or year <= 0:
            raise ValueError("An argument is invalid")
        return self.__retrieve_holidays(country_code, year)




import base64, datetime, hashlib, hmac, urllib
from datetime import datetime
from typing import Dict

class Authentication:
    def __init__(self, service: str, access_key: str, secret_key: str):
        self.__service: str = service
        self.__access_key: str = access_key
        self.__secret_key: str = secret_key

    def get_authentication_args(self, seed: Dict[str, object] = None) -> Dict[str, object]:
        timestamp = datetime.utcnow().isoformat()
        message = self.__access_key + self.__service + timestamp;
        digester = hmac.new(bytes(self.__secret_key, encoding="utf8"), bytes(message, encoding="utf8"), hashlib.sha1)

        args: Dict[str, object] = seed if seed is not None else {}
        args["accesskey"] = self.__access_key
        args["timestamp"] = timestamp
        args["signature"] = base64.b64encode(digester.digest())

        return args




import xml.etree.ElementTree as ET

class Country:
    '''
    A class used for country data

    ...

    Attributes
    ----------
    id : str
        The ISO 3166-1-alpha-2 country code.
    name : str
        Full name of the country.
    '''

    def __init__(self, node: ET.Element):
        self.id: str = None
        self.name: str = None

        if node is not None and node.text:
            self.name = node.text
            cid = node.get("id")
            if cid:
                self.id = cid




from libtad.base_service import BaseService
from libtad.datatypes.astro import AstronomyEventClass, AstronomyObjectType, AstronomyLocation
from libtad.datatypes.places import LocationId
from libtad.datatypes.time import TADDateTime
from libtad.common.exceptions import QueriedDateOutOfRangeException
from libtad.common import XmlUtils
import libtad.constants as Constants
from typing import List, Dict
import xml.etree.ElementTree as ET
from urllib.parse import urlencode
from urllib.request import urlopen, Request
from ssl import SSLContext

class AstronomyService(BaseService):
    """
    The astronomy service can be used retrieve the sunrise, sunset, moonrise, 
    moonset, solar noon and twilight times for all locations in our database.

    ...

    Attributes
    ----------
    types : AstronomyEventClass
        Selection of which astronomical events you are interested in.
        To combine multiple classes, use the binary OR operator.

        Example:
        ``service.Types = AstronomyEventClass.CivilTwilight | AstronomyEventClass.NauticalTwilight``
    include_coordinates : bool
        Return longitude and latitude for the geo object.
    include_isotime : bool
        Adds timestamps (local time) to all events.
    include_utctime : bool
        Adds UTC timestamps to all events.
    radius : int
        Search radius for translating coordinates (parameter placeid) to locations. 
        Coordinates that could not be translated will yield results for the actual 
        geographical position – if you would like to query for times at an exact 
        location, specify a radius of zero (0).

        The radius in kilometers. Default is infinite, but only locations within 
        the same country and time zone are considered.
    language : list of str
        The preferred language(s) for the texts. An error will be raised if the
        language code cannot be recognized. In case the text for a specific event
        cannot be retrieved in any of the requested languages it will be returned
        in English instead.

        It is possible to query all available languages in a single request by
        specifying a primary text language, followed by the special value _all_.
        This value can not be specified as the first language.

        Example:
        ``service.language = ["de"]``

        Example:
        ``service.language.append("de")``
    
    Methods
    -------
    get_astronomical_info(object_type, place_id, start_date, end_date=None)
        Gets the specified object type for a specified place by start date.        
    """

    def __init__(self, access_key: str, secret_key: str):
        """
        Parameters
        ----------
        access_key : str
            Access key.
        secret_key : str
            Secret key.
        """

        super().__init__(access_key, secret_key, "astronomy")
        self.types: AstronomyEventClass = AstronomyEventClass(0)
        self.include_coordinates: bool = True
        self.include_isotime: bool = False
        self.include_utctime: bool = False
        self.radius: int = None

    def get_astronomical_info(self, object_type: AstronomyObjectType, place_id: LocationId, start_date: TADDateTime, end_date: TADDateTime = None) -> List[AstronomyLocation]:
        """
        Gets the specified object type for a specified place by start date.

        Parameters
        ----------
        object_type : AstronomyObjectType
            The astronomical object type.
        place_id : LocationId
            Place identifier.
        start_date : TADDateTime
            Start date.
        end_date : TADDateTime, optional
            End date.
        
        Returns
        -------
        astronomy_locations : list of AstronomyLocation
            A list of astronomical information.

        """

        if type(place_id) is not LocationId or start_date.year == 0 or not object_type:
            raise ValueError("An argument is invalid")

        ID = str(place_id)
        if not ID:
            raise ValueError("An argument is invalid")

        if end_date and end_date < start_date:
            QueriedDateOutOfRangeException("End date cannot be before start date")

        args: Dict[str, object] = self._authentication_options.copy()
        args["placeid"] = place_id
        args["object"] = object_type.name.lower()
        args["startdt"] = str(start_date)
        if end_date:
            args["enddt"] = str(end_date)

        return self.__retrieve_astronomical_info(args)

    def __retrieve_astronomical_info(self, args: Dict[str, object]) -> List[AstronomyLocation]:
        arguments: Dict[str, object] = self.__get_optional_arguments(args)
        url: str = Constants.ENTRYPOINT + "/" + self._service_name + "?" + urlencode(arguments)
        req = Request(
                url,
                headers = {"User-Agent": "libtad-py"}
                )

        with urlopen(req, context=SSLContext()) as f:
            result: str = f.read().decode("utf-8")

        return self.__from_xml(result)
        
    def __get_optional_arguments(self, args: Dict[str, object]) -> Dict[str, object]:
        optional_args: Dict[str, str] = args
        types: int = self.__get_astronomy_event_types()

        optional_args["geo"] = int(self.include_coordinates)
        optional_args["isotime"] = int(self.include_isotime)
        optional_args["lang"] = ",".join(self.language)
        optional_args["utctime"] = int(self.include_utctime)
        optional_args["out"] = Constants.DEFAULTRETURNFORMAT
        optional_args["verbosetime"] = str(Constants.DEFAULTVERBOSETIMEVALUE)
        optional_args["version"] = str(self._version)

        if self.radius is not None:
            optional_args["radius"] = str(self.radius)
        if types:
            optional_args["types"] = types

        return optional_args

    def __from_xml(self, result: str) -> List[AstronomyLocation]:
        XmlUtils.check_for_errors(result)
        xml: ET.Element = ET.fromstring(result)
        return [AstronomyLocation(location) for location in xml.findall("location")]

    def __get_astronomy_event_types(self) -> str:
        included_strings: List[str] = []
        for astro_item in AstronomyEventClass:
            if astro_item & self.types:
                included_strings.append(AstronomyEventClass.resolve(astro_item)[1])
        included: str = ",".join(included_strings)
        return included




from libtad.datatypes.time import TADDateTimeOffset
from .moon_phase import MoonPhase
import xml.etree.ElementTree as ET

class AstronomyCurrent:
    """
    A class used to store current data for an astronomy object.

    ...

    Attributes
    ----------
    utctime: TADDateTimeOffset
        UTC time stamp for the data in ISO 8601 format.
    isotime : TADDateTimeOffset
        Local time stamp for the data in ISO 8601 format (including UTC offset).
    azimuth : float 
        Horizontal direction of the astronomical object at set/rise time 
        (referring to true north). North is 0 degrees, east is 90 degrees, south 
        is 180 degrees and west is 270 degrees. 
    altitude : float 
        Altitude of the center of the queried astronomical object above an 
        ideal horizon. 
    distance : float 
        Distance of the earth's center to the center of the queried astronomical 
        object in kilometers. 
    illuminated : float 
        The fraction of the Moon's surface illuminated by the Sun's rays as seen 
        from the selected location. Only available for the moon object. 
    posangle : float 
        The counterclockwise angle of the midpoint of the Moon's bright limb as 
        seen from the selected location. Only available for the moon object. 
    moonphase : MoonPhase
        The current phase of the moon. Only available for the moon object. 
    """

    def __init__(self, node: ET.Element):
        self.utctime: TADDateTimeOffset = None
        self.isotime: TADDateTimeOffset = None
        self.azimuth: float = None
        self.altitude: float = None
        self.distance: float = None
        self.illuminated: float = None
        self.posangle: float = None
        self.moonphase: MoonPhase = MoonPhase.NotRequested

        utctime = node.get("utctime")
        isotime = node.get("isotime")
        azimuth = node.find("azimuth")
        altitude = node.find("altitude")
        distance = node.find("distance")
        illuminated = node.find("illuminated")
        posangle = node.find("posangle")
        moonphase = node.find("moonphase")
        
        if utctime:
            self.utctime = TADDateTimeOffset._parse(utctime)

        if isotime:
            self.isotime = TADDateTimeOffset._parse(isotime)

        if azimuth is not None:
            self.azimuth = float(azimuth.text)

        if altitude is not None:
            self.altitude = float(altitude.text)

        if distance is not None:
            self.distance = float(distance.text)

        if illuminated is not None:
            self.illuminated = float(illuminated.text)

        if posangle is not None:
            self.posangle = float(posangle.text)

        if moonphase is not None:
            self.moonphase = MoonPhase._parse(moonphase.text)




from typing import Union, Tuple
from enum import Flag

class HolidayType(Flag):
    '''
        An enum class with simple bitflags so the user can easily put together a set of types
        which should be included or excluded in the retrieval of holidays.
    '''

    All = 1
    Default = 1 << 1
    DefaultForCountry = 1 << 2
    Observances = 1 << 3
    Federal = 1 << 4
    FederalLocal = 1 << 5
    Local = 1 << 6
    Flagdays = 1 << 7
    LocalObservances = 1 << 8
    ImportantObservances = 1 << 9
    CommonObservances = 1 << 10
    OtherObservances = 1 << 11
    Weekdays = 1 << 12
    Buddhism = 1 << 13
    Hebrew = 1 << 14
    Hinduism = 1 << 15
    Muslim = 1 << 16
    Orthodox = 1 << 17
    Seasons = 1 << 18
    TimezoneEvents = 1 << 19
    UnitedNations = 1 << 20
    WorldWideObservances = 1 << 21
    Christian = 1 << 22
    Defacto = 1 << 23
    Religious = 1 << 24
    Halfday = 1 << 25
    Optional = 1 << 26
    Otherreligion = 1 << 27
    Sport = 1 << 28
    Fun = 1 << 29

    @staticmethod
    def resolve(holiday_type: Union[int, str, "HolidayType"]) -> Tuple[int, str, "HolidayType"]:
        tup: Tuple[Tuple[int, str, HolidayType], ...] = (
                    (1, "all", HolidayType.All),
                    (1 << 1, "default", HolidayType.Default),
                    (1 << 2, "countrydefault", HolidayType.DefaultForCountry),
                    (1 << 3, "obs", HolidayType.Observances),
                    (1 << 4, "federal", HolidayType.Federal),
                    (1 << 5, "federallocal", HolidayType.FederalLocal),
                    (1 << 6, "local", HolidayType.Local),
                    (1 << 7, "flagday", HolidayType.Flagdays),
                    (1 << 8, "local2", HolidayType.LocalObservances),
                    (1 << 9, "obs1", HolidayType.ImportantObservances),
                    (1 << 10, "obs2", HolidayType.CommonObservances),
                    (1 << 11, "obs3", HolidayType.OtherObservances),
                    (1 << 12, "weekday", HolidayType.Weekdays),
                    (1 << 13, "buddhism", HolidayType.Buddhism),
                    (1 << 14, "hebrew", HolidayType.Hebrew),
                    (1 << 15, "hinduism", HolidayType.Hinduism),
                    (1 << 16, "muslim", HolidayType.Muslim),
                    (1 << 17, "orthodox", HolidayType.Orthodox),
                    (1 << 18, "seasons", HolidayType.Seasons),
                    (1 << 19, "tz", HolidayType.TimezoneEvents),
                    (1 << 20, "un", HolidayType.UnitedNations),
                    (1 << 21, "world", HolidayType.WorldWideObservances),
                    (1 << 22, "christian", HolidayType.Christian),
                    (1 << 23, "defacto", HolidayType.Defacto),
                    (1 << 24, "religious", HolidayType.Religious),
                    (1 << 25, "halfday", HolidayType.Halfday),
                    (1 << 26, "optional", HolidayType.Optional),
                    (1 << 27, "otherreligion", HolidayType.Otherreligion),
                    (1 << 28, "sport", HolidayType.Sport),
                    (1 << 29, "fun", HolidayType.Fun)
                    )

        if type(holiday_type) is int:
            index = 0
        elif type(holiday_type) is str:
            index = 1
        elif type(holiday_type) is HolidayType:
            index = 2
        else:
            return None

        for hol_entry in tup:
            if holiday_type == hol_entry[index]:
                return hol_entry




from enum import Flag

class DSTSpecialType(Flag):
    """
        Indicates if the region does not observe DST at all, or is on DST all year long.
    """

    NoDaylightSavingTime = 1
    DaylightSavingTimeAllYear = 1 << 1

    @staticmethod
    def resolve(type_code: str) -> "DSTSpecialType":
        if type_code == "nodst":
            return DSTSpecialType.NoDaylightSavingTime
        elif type_code == "allyear":
            return DSTSpecialType.DaylightSavingTimeAllYear
        else:
            raise ValueError("type_code does not conform to enum DSTSpecialType")




from .country import Country
from .coordinates import Coordinates
import xml.etree.ElementTree as ET

class Geo:
    """
    The geographical information type.

    ...

    Attributes
    ----------
    name : str
        The name of the location.
    state : str
        The state of the location within the country (only if applicable).
    country : Country
        Country of the location.
    coordinates : Coordinates
        Geographical coordinates of the location.
    """

    def __init__(self, node: ET.Element):
        self.name: str = None
        self.state: str = None
        self.country: Country = None
        self.coordinates: Coordinates = None

        latitude = node.find("latitude")
        longitude = node.find("longitude")
        name = node.find("name")
        state = node.find("state")
        country = node.find("country")

        if all(l is not None and l.text is not None for l in (latitude, longitude)):
            self.coordinates = Coordinates(float(latitude.text), float(longitude.text))

        if name is not None:
            self.name = name.text

        if state is not None:
            self.state = state.text

        if country is not None:
            self.country = Country(country)




from .business_filter_method import BusinessFilterMethod
from libtad.datatypes.holidays import Holiday
import xml.etree.ElementTree as ET
from typing import List

class BusinessHolidayType:
    def __init__(self, node: ET.Element):
        """
        Holidays which occur in a given period.

        ...

        Attributes
        ----------
        type : BusinessFilterMethod
            Specifies whether or not the holidays in the result array were 
            included or excluded when queried.
        count : int
            The number of holidays in the results.
        list : list of Holiday
            Holidays which occur in the requested period.
        """

        self.type: BusinessFilterMethod = None
        self.count: int = 0
        self.list: List[Holiday] = []

        _type = node.get("type")
        count = node.get("count")
        _list = node.find("list")

        if _type:
            try:
                self.type = BusinessFilterMethod[_type.capitalize()]
            except KeyError:
                pass

        if count:
            try:
                self.count = int(count)
            except ValueError:
                pass

        if _list is not None:
            self.list = [Holiday(holiday) for holiday in _list.findall("holiday")]




__all__ = [
    "BusinessDaysOperatorType",
    "BusinessDaysFilterType",
    "BusinessDates",
    "BusinessFilterMethod",
    "BusinessHolidayType",
    "PeriodType",
    "WeekdaysType",
]

from libtad.datatypes.business.business_days_operator_type import BusinessDaysOperatorType
from libtad.datatypes.business.business_days_filter_type import BusinessDaysFilterType
from libtad.datatypes.business.business_dates import BusinessDates
from libtad.datatypes.business.business_filter_method import BusinessFilterMethod
from libtad.datatypes.business.business_holiday_type import BusinessHolidayType
from libtad.datatypes.business.period_type import PeriodType
from libtad.datatypes.business.weekdays_type import WeekdaysType

def __dir__():
    return __all__


from libtad.base_service import BaseService
from libtad.datatypes.places import LocationId, Location
from libtad.common import XmlUtils
import libtad.constants as Constants
from typing import List, Dict, Union
from urllib.parse import urlencode
from urllib.request import urlopen, Request
from ssl import SSLContext
import xml.etree.ElementTree as ET

class TimeService(BaseService):
    """
    The timeservice service can be used to retrieve the current time in one 
    or more places. Additionally, information about time zones and related 
    changes and the time of sunrise and sunset can be queried.

    ...

    Attributes
    ----------
    radius : int
        Search radius for translating coordinates (parameter placeid) to locations. 
        Coordinates that could not be translated will yield results for the actual 
        geographical position.

        The radius in kilometers. Default is infinite, but only locations within 
        the same country and time zone are considered.
    include_coordinates : bool
        Return coordinates for the Geography object.
    include_sunrise_and_sunset : bool
        Controls if the astronomy element with information about sunrise and 
        sunset shall be added to the result.
    include_current_time_to_location : bool
        Adds current time under the location object.
    include_list_of_time_changes : bool
        Add a list of time changes during the year to the location object. This 
        listing e.g. shows changes caused by daylight savings time.
    include_timezone_information : bool
        Add timezone information under the time object.

    Methods
    -------
    current_time_for_place(place_id)
        Retrieves the current time for place by ID.
    """

    def __init__(self, access_key: str, secret_key: str):
        """
        Parameters
        ----------
        access_key : str
            Access key.
        secret_key : str
            Secret key.
        """

        super().__init__(access_key, secret_key, "timeservice")
        self.radius: int = None
        self.include_coordinates: bool = True
        self.include_sunrise_and_sunset: bool = True
        self.include_current_time_to_location: bool = True
        self.include_list_of_time_changes: bool = True
        self.include_timezone_information: bool = True

    def current_time_for_place(self, place_id: Union[LocationId, List[LocationId]]) -> List[Location]:
        """
        Retrieves the current time for place by ID.

        Parameters
        ----------
        place_id : LocationId or list of LocationId
            Place identifier.

        Returns
        -------
        location_times : list of Location
            The current time for place.

        """

        place_id_str = ""
        if isinstance(place_id, LocationId):
            place_id_str = str(place_id)
        elif isinstance(place_id, list) and all(isinstance(pid, LocationId) for pid in place_id):
            place_id_str = ",".join([str(pid) for pid in place_id])

        if not place_id_str:
            raise ValueError("An argument is invalid")

        args: Dict[str, object] = self._authentication_options.copy()
        args["placeid"] = place_id_str

        return self.__retrieve_current_time(args)

    def __retrieve_current_time(self, args: Dict[str, object]) -> List[Location]:
        arguments: Dict[str, object] = self.__get_optional_arguments(args)
        url: str = Constants.ENTRYPOINT + "/" + self._service_name + "?" + urlencode(arguments)
        req = Request(
                url,
                headers = { "User-Agent": "libtad-py"}
                )

        with urlopen(req, context=SSLContext()) as f:
            result: str = f.read().decode("utf-8")

        return self.__from_xml(result)

    def __get_optional_arguments(self, args: Dict[str, object]) -> Dict[str, object]:
        args["geo"] = int(self.include_coordinates)
        args["lang"] = ",".join(self.language)
        args["sun"] = int(self.include_sunrise_and_sunset)
        args["time"] = int(self.include_current_time_to_location)
        args["timechanges"] = int(self.include_list_of_time_changes)
        args["tz"] = int(self.include_timezone_information)
        args["out"] = Constants.DEFAULTRETURNFORMAT
        args["verbosetime"] = str(Constants.DEFAULTVERBOSETIMEVALUE)
        args["version"] = str(self._version)

        if self.radius is not None:
            args["radius"] = self.radius

        return args
    
    def __from_xml(self, result: str) -> List[Location]:
        XmlUtils.check_for_errors(result)
        xml: ET.Element = ET.fromstring(result)
        return [Location(loc) for loc in xml.findall("location")]




from .tad_time_span import TADTimeSpan
import xml.etree.ElementTree as ET

class TADTimezone:
    '''
    A class used to store information about timezones.

    ...

    Attributes
    ----------
    abbreviation : str
        Abbreviated timezone name.

        Example: LHDT
    name : str
        Full timezone name.

        Example: Lord Howe Daylight Time
    offset : TADTimeSpan
        The timezone offset (from UTC) as TimeSpan.
    basic_offset : int
        Basic timezone offset (without DST) in seconds.
    dst_offset : int
        DST component of timezone offset in seconds.
    total_offset : int
        Total offset from UTC in seconds.
    '''

    def __init__(self, node: ET.Element):
        self.abbreviation: str = ""
        self.name: str = ""
        self.offset: TADTimeSpan = None
        self.basic_offset: int = None
        self.dst_offset: int = None
        self.total_offset: int = None

        zoneabb_node = node.find("zoneabb")
        name_node = node.find("zonename")
        zoneoffset_node = node.find("zoneoffset")
        zonedst_node = node.find("zonedst")
        totaloffset_node = node.find("zonetotaloffset")
        offset = node.get("offset")

        if zoneabb_node is not None:
            self.abbreviation = zoneabb_node.text

        if name_node is not None:
            self.name = name_node.text

        if zoneoffset_node is not None:
            self.basic_offset = int(zoneoffset_node.text)

        if zonedst_node is not None:
            self.dst_offset = int(zonedst_node.text)

        if totaloffset_node is not None:
            self.total_offset = int(totaloffset_node.text)

        if offset:
            self.offset = TADTimeSpan._parse(offset)




from setuptools import setup, find_packages

PACKAGES = find_packages()

setup(name='libtad',
      version='1.1.1',
      description='Python library for accessing Time and Date APIs',
      author='Time and Date',
      url='https://github.com/timeanddate/libtad-python/',
      packages=PACKAGES,
      )



from libtad.base_service import BaseService
from libtad.datatypes.places import LocationId
from libtad.datatypes.time import TADDateTime
from libtad.datatypes.converted_times import ConvertedTimes
from libtad.common import XmlUtils
import libtad.constants as Constants
from typing import List, Dict, Union
import xml.etree.ElementTree as ET
from urllib.parse import urlencode
from urllib.request import urlopen, Request
from ssl import SSLContext

class ConvertTimeService(BaseService):
    """
    The converttime service can be used to convert any time from UTC or any of 
    the supported locations to any other of the supported locations.
    ...

    Attributes
    ----------
    radius : int
        Search radius for translating coordinates (parameters fromid and toid)
        to locations. Coordinates that could not be translated will yield results
        for the actual geographical position.
    include_time_changes : bool
        Add a list of time changes during the year to the location object. This listing
        e.g. shows changes caused by daylight savings time.
    include_timezone_information : bool
        Add timezone information under the time object.
    language : list of str
        The preferred language(s) for the texts. An error will be raised if the
        language code cannot be recognized. In case the text for a specific event
        cannot be retrieved in any of the requested languages it will be returned
        in English instead.

        It is possible to query all available languages in a single request by
        specifying a primary text language, followed by the special value _all_.
        This value can not be specified as the first language.

        Example:
        ``service.language = ["de"]``
 
        Example:
        ``service.language.append("de")``

    Methods
    -------
    convert_time(from_id, time, to_ids=None)
        Converts the time by using a LocationId, a ISO-string and optionally a list 
        of IDs to convert to.
    """

    def __init__(self, access_key: str, secret_key: str):
        """
        Parameters
        ----------
        access_key : str
            Access key.
        secret_key : str
            Secret key.
        """

        super().__init__(access_key, secret_key, "converttime")
        self.radius: int = None
        self.include_time_changes: bool = True
        self.include_timezone_information: bool = True
        
    def convert_time(self, from_id: LocationId, time: Union[str, TADDateTime], to_ids : List[LocationId] = None) -> ConvertedTimes:
        """
        Converts the time by using a LocationId, a ISO-string and optionally a list
        of IDs to convert to.

        Parameters
        ----------
        from_id : LocationId
            The places identifier.
        time : str or TADDateTime
            ISO 8601-formatted string or TADDateTime object.
        to_ids : list of LocationId, optional
            The place IDs to convert to.
        
        Returns
        -------
        converted_time : ConvertedTimes
            An object containing UTC time stamp of requested time and a list of time 
            information for the locations in the request.

        """

        if (not isinstance(from_id, LocationId)
                or not (isinstance(time, TADDateTime) or isinstance(time, str))
                or not (not to_ids or (isinstance(to_ids, list) and all(isinstance(to_id, LocationId) for to_id in to_ids)))):
            raise ValueError("An argument is invalid")

        from_id_str = str(from_id)
        if isinstance(time, TADDateTime):
            time_str = str(time)

        if not from_id_str or not time_str:
            raise ValueError("An argument is invalid")

        args: Dict[str, object] = self._authentication_options.copy()
        args["fromid"] = from_id_str
        args["iso"] = time_str

        if to_ids:
            args["toid"] = ",".join([str(to_id) for to_id in to_ids])

        return self.__retrieve_converted_times(args)

    def __retrieve_converted_times(self, args: Dict[str, object]) -> ConvertedTimes:
        arguments: Dict[str, object] = self.__get_optional_arguments(args)
        url: str = Constants.ENTRYPOINT + "/" + self._service_name + "?" + urlencode(arguments)
        req = Request(
                url,
                headers = { "User-Agent": "libtad-py"}
                )

        with urlopen(req, context=SSLContext()) as f:
            result: str = f.read().decode("utf-8")

        return self.__from_xml(result)

    def __get_optional_arguments(self, args: Dict[str, object]) -> Dict[str, object]:
        optional_args: Dict[str, str] = args

        optional_args["timechanges"] = int(self.include_time_changes)
        optional_args["tz"] = int(self.include_timezone_information)
        optional_args["lang"] = ",".join(self.language)
        optional_args["out"] = Constants.DEFAULTRETURNFORMAT
        optional_args["verbosetime"] = str(Constants.DEFAULTVERBOSETIMEVALUE)
        optional_args["version"] = str(self._version)

        if self.radius is not None:
            optional_args["radius"] = str(self.radius)

        return optional_args

    def __from_xml(self, result: str) -> ConvertedTimes:
        XmlUtils.check_for_errors(result)
        xml: ET.Element = ET.fromstring(result)
        return ConvertedTimes(xml.find("utc"), [location for location in xml.findall("location")])




from libtad.datatypes.time import TADDateTime, TADTimeSpan
from .moon_phase import MoonPhase
from .astronomy_day_event import AstronomyDayEvent
import xml.etree.ElementTree as ET
from typing import List

class AstronomyDay:
    """
    A class used for storing astronomy object information for a given day.

    ...

    Attributes
    ----------
    date : TADDateTime
        Date for the current information.
    day_length : TADTimeSpan
        Length of this day (time between sunrise and sunset). If the sun is not
        up on this day, 00:00 will reported. If the sun does not set on this day,
        the value will read 24:00.
    moonphase : MoonPhase
        Moon phase for the day. Only if requested.
    events : list of AstronomyDayEvent
        Lists all events during the day.

    """

    def __init__(self, node: ET.Element):
        self.date: TADDateTime = None
        self.day_length: TADTimeSpan = None
        self.moonphase: MoonPhase = MoonPhase.NotRequested
        self.events: List[AstronomyDayEvent] = []

        date = node.get("date")
        day_length = node.get("daylength")
        moonphase = node.get("moonphase")
        events = node.findall("event")

        if date:
            self.date = TADDateTime._parse(date)

        if day_length:
            self.day_length = TADTimeSpan._parse(day_length)

        if moonphase:
            phase = MoonPhase._parse(moonphase)
            if phase:
                self.moonphase: MoonPhase = phase

        if events:
            self.events = [AstronomyDayEvent(event) for event in events]




__all__ = [
    "HolidayType",
    "Holiday",
    "HolidayState"
]

from libtad.datatypes.holidays.holiday_type import HolidayType
from libtad.datatypes.holidays.holiday import Holiday
from libtad.datatypes.holidays.holiday_state import HolidayState

def __dir__():
    return __all__


class ServerSideException(Exception):
    def __init__(self, message: str):
        super().__init__(f"An error occured on server-side: {message}")

class MalformedXMLException(Exception):
    def __init__(self, message: str, ignore_default=False):
        super().__init__(("The XML returned from Time and Date contained an unsupported name: " if not ignore_default else "") + message)

class QueriedDateOutOfRangeException(Exception):
    pass




__all__ = [
    "AstronomyDay",
    "AstronomyDayEvent",
    "AstronomyEventCode",
    "AstronomyObjectType",
    "AstronomyCurrent",
    "AstronomyObjectDetails",
    "AstronomyLocation",
    "Astronomy",
    "AstronomyEventClass",
    "MoonPhase",
    "AstronomyEvent",
    "AstronomyEventType",
    "AstronomySpecialType",
    "AstronomySpecial",
]

from libtad.datatypes.astro.astronomy_day import AstronomyDay
from libtad.datatypes.astro.astronomy_day_event import AstronomyDayEvent
from libtad.datatypes.astro.astronomy_event_code import AstronomyEventCode
from libtad.datatypes.astro.astronomy_object_type import AstronomyObjectType
from libtad.datatypes.astro.astronomy_current import AstronomyCurrent
from libtad.datatypes.astro.astronomy_object_details import AstronomyObjectDetails
from libtad.datatypes.astro.astronomy_location import AstronomyLocation
from libtad.datatypes.astro.astronomy import Astronomy
from libtad.datatypes.astro.astronomy_event_class import AstronomyEventClass
from libtad.datatypes.astro.moon_phase import MoonPhase
from libtad.datatypes.astro.astronomy_event import AstronomyEvent
from libtad.datatypes.astro.astronomy_event_type import AstronomyEventType
from libtad.datatypes.astro.astronomy_special_type import AstronomySpecialType
from libtad.datatypes.astro.astronomy_special import AstronomySpecial

def __dir__():
    return __all__



from .tad_date_time import TADDateTime
from .tad_time_span import TADTimeSpan

class TADDateTimeOffset:
    """
    A class used to store datetime with offsets.
    
    ...

    Attributes
    ----------
    date_time : TADDateTime
        Storage of time and date.
    time_span : TADTimeSpan
        Offset.
    """

    def __init__(self, date_time: TADDateTime, time_span: TADTimeSpan):
        self.date_time: TADDateTime = date_time
        self.time_span: TADTimeSpan = time_span

    @staticmethod
    def _parse(fmtstr: str) -> "TADDateTimeOffset":
        strlist = fmtstr.split("T")
        if len(strlist) != 2:
            raise ValueError("Cannot parse DateTimeOffset string")

        offset_prefix = None
        if "-" in strlist[1]:
            offset_prefix = "-"
        elif "+" in strlist[1]:
            offset_prefix = "+"

        if offset_prefix:
            last = strlist[1].split(offset_prefix)
            if len(last) != 2:
                raise ValueError("Cannot parse DateTimeOffset string")
            strlist = [strlist[0], last[0], last[1]]

        try:
            datetime = TADDateTime._parse("T".join(strlist[:2]))
            offset_list = [int(num) * int(offset_prefix + "1") for num in strlist[2].split(":")] if offset_prefix else [0, 0]
            if len(offset_list) != 2:
                raise ValueError
        except ValueError:
            raise ValueError("Cannot parse DateTimeOffset string")

        timespan = TADTimeSpan(hours=offset_list[0], minutes=offset_list[1])
        return TADDateTimeOffset(datetime, timespan)




from . import TADDateTime
import xml.etree.ElementTree as ET

class TimeChange:
    """
    A class used to store information about time change.

    ...

    Attributes
    ----------
    new_daylight_saving_time : int
        New DST offset in seconds. Value will be null if there is
        no DST for this location.
    new_timezone_offset : int
        New timezone offset to UTC in seconds if there is a timezone
        change for this place. Otherwise the value will be null.
        Time zones changes happen only very rarely, so the field will
        be null on most occasions.
    new_total_offset : int
        New total offset to UTC in seconds.
    utc_time : TADDateTime
        The UTC time of the transition.
    old_local_time : TADDateTime
        The old local time before the transition.
    new_local_time : TADDateTime
        The new local time after the transition.
    """

    def __init__(self, node: ET.Element):
        self.new_daylight_saving_time: int = None
        self.new_timezone_offset: int = None
        self.new_total_offset: int = None
        self.utc_time: TADDateTime = None
        self.old_local_time: TADDateTime = None
        self.new_local_time: TADDateTime = None

        newdst = node.get("newdst")
        newzone = node.get("newzone")
        newoffset = node.get("newoffset")
        utctime = node.get("utctime")
        oldlocal = node.get("oldlocaltime")
        newlocal = node.get("newlocaltime")

        if newdst:
            try:
                self.new_daylight_saving_time = int(newdst)
            except ValueError:
                pass

        if newzone:
            try:
                self.new_timezone_offset = int(newzone)
            except ValueError:
                pass

        if newoffset:
            try:
                self.new_total_offset = int(newoffset)
            except ValueError:
                pass

        if utctime:
            self.utc_time = TADDateTime._parse(utctime)

        if oldlocal:
            self.old_local_time = TADDateTime._parse(oldlocal)

        if newlocal:
            self.new_local_time = TADDateTime._parse(newlocal)




from libtad.base_service import BaseService
from libtad.datatypes.places import Place
from libtad.common import XmlUtils
import libtad.constants as Constants
import xml.etree.ElementTree as ET
from urllib.parse import ParseResult, urlunparse, urlencode
from urllib.request import urlopen, Request
from ssl import SSLContext
from typing import List, Dict


class PlacesService(BaseService):
    """
    The places service can be used to retrieve the list of supported places. 

    ...

    Attributes
    ----------
    include_coordinates : bool
        Return coordinates for the Geography object.

    Methods
    -------
    get_places()
        Gets list of supported places.
    """

    def __init__(self, access_key: str, secret_key: str):
        """
        Parameters
        ----------
        access_key : str
            Access key.
        secret_key : str
            Secret key.
        """

        super().__init__(access_key, secret_key, "places")
        self.include_coordinates: bool = True

    def get_places(self) -> List[Place]:
        """
        Gets list of supported places.

        Returns
        -------
        places : list of Place
            List of all currently known places, their identifiers and their 
            geographical location (if requested).

        """

        args = self.__get_arguments()
        url: str = Constants.ENTRYPOINT + "/" + self._service_name + "?" + urlencode(args)
        req = Request(
                url,
                headers = { "User-Agent": "libtad-py"}
                )

        with urlopen(req, context=SSLContext()) as f:
            result: str = f.read().decode("utf-8")

        return self.__from_xml(result)


    def __get_arguments(self) -> Dict[str, object]:
        args: Dict[str, object] = self._authentication_options.copy()
        args["lang"] = ",".join(self.language)
        args["geo"] = int(self.include_coordinates)
        args["version"] = str(self._version)
        args["out"] = Constants.DEFAULTRETURNFORMAT
        args["verbosetime"] = str(Constants.DEFAULTVERBOSETIMEVALUE)

        return args

    def __from_xml(self, result: str) -> List[Place]:
        XmlUtils.check_for_errors(result)
        xml: ET.Element = ET.fromstring(result)
        places = xml.find("places")
        return [Place(place_node) for place_node in places.findall("place")]




from libtad.base_service import BaseService
from libtad.datatypes.dst import DST
import libtad.constants as Constants
from libtad.common import XmlUtils
from typing import List, Dict
import xml.etree.ElementTree as ET
from urllib.parse import urlencode
from urllib.request import urlopen, Request
from ssl import SSLContext

class DSTService(BaseService):
    """
    The dstlist service can be used to obtain data about time zones for all
    supported countries in our database. This includes the start and end
    date of daylight savings time, and UTC offset for the time zones.

    The resulting data is aggregated on country and time zone level. 
    By default, only information from countries which actually observe DST 
    is returned without listing the individually affected locations – see 
    the parameters listplaces and onlydst to change this behavior.

    ...

    Attributes
    ----------
    include_time_changes : bool
        Add a list of time changes during the year to the dstentry object. 
        This listing e.g. shows changes caused by daylight savings time.

    include_only_dst_countries : bool
        Return only countries which actually observe DST in the queried year. 
        Other countries will be suppressed.

    include_places_for_every_country : bool
        For every time zone/country, list the individual places that belong to
        each record.

    language : list of str
        The preferred language(s) for the texts. An error will be raised if the
        language code cannot be recognized. In case the text for a specific event
        cannot be retrieved in any of the requested languages it will be returned
        in English instead.

        It is possible to query all available languages in a single request by
        specifying a primary text language, followed by the special value _all_.
        This value can not be specified as the first language.

        Example:
        ``service.language = ["de"]``

        Example:
        ``service.language.append("de")``

    Methods
    -------
    get_daylight_saving_time(country_code=None, year=None)
        Gets the daylight saving time by country and year.
    """

    def __init__(self, access_key: str, secret_key: str):
        """
        Parameters
        ----------
        access_key : str
            Access key.
        secret_key : str
            Secret key.
        """

        super().__init__(access_key, secret_key, "dstlist")
        self.include_time_changes: bool = False
        self.include_only_dst_countries: bool = True
        self.include_places_for_every_country: bool = True

    def get_daylight_saving_time(self, country_code: str = None, year: int = None) -> List[DST]:
        """
        Gets the daylight saving time by country and year.

        If `country_code` is unspecified, information for all countries will be
        returned. If `year` is unspecified, current year is used.

        Parameters
        ----------
        country_code : str, optional
            ISO3166-1-alpha-2 Country Code.
            Uses all countries by default.
        year : int, optional
            Year.
            Uses the current year by default.

        Returns
        -------
        dst_list : list of DST
            The DST information for each country or region.

        """

        args: Dict[str, object] = {}
        
        if country_code and type(country_code) is str:
            args["country"] = country_code
        if year and type(year) is int:
            if year <= 0:
                raise ValueError("An argument is invalid")
            args["year"] = str(year)
            self.include_only_dst_countries = False

        return self.__retrieve_dst_list(args)

    def __retrieve_dst_list(self, args: Dict[str, object]) -> List[DST]:
        args.update(self.__get_arguments())
        url: str = Constants.ENTRYPOINT + "/" + self._service_name + "?" + urlencode(args)
        req = Request(
                url,
                headers = { "User-Agent": "libtad-py"}
                )

        with urlopen(req, context=SSLContext()) as f:
            result: str = f.read().decode("utf-8")

        return self.__from_xml(result)

    def __get_arguments(self) -> Dict[str, object]:
        args: Dict[str, object] = self._authentication_options.copy()
        args["lang"] = ",".join(self.language)
        args["out"] = Constants.DEFAULTRETURNFORMAT
        args["verbosetime"] = str(Constants.DEFAULTVERBOSETIMEVALUE)
        args["version"] = str(self._version)
        args["timechanges"] = int(self.include_time_changes)
        args["onlydst"] = int(self.include_only_dst_countries)
        args["listplaces"] = int(self.include_places_for_every_country)
        return args

    def __from_xml(self, result: str) -> List[DST]:
        XmlUtils.check_for_errors(result)
        xml: ET.Element = ET.fromstring(result)
        dstlist: ET.Element = xml.find("dstlist")
        return [DST(dstentry) for dstentry in dstlist.findall("dstentry")]




class Coordinates:
    def __init__(self, latitude: float, longitude: float):
        self.latitude: float = latitude
        self.longitude: float = longitude

    def __str__(self) -> str:
        coords = ""
        lat = self.latitude
        long_ = self.longitude

        if lat >= 0:
            coords += "+"
        coords += str(lat)

        if long_ >= 0:
            coords += "+"
        coords += str(long_)

        return coords

